import typing


def format_scheme_value(
    value: typing.Any, force_quotes: bool = False, verbatim: bool = False
) -> str:
    r"""
    Formats ``value`` in a scheme-like way.

    ..  container:: example

        Some basic values:

        >>> abjad.fsv.format_scheme_value(1)
        '1'

        >>> abjad.fsv.format_scheme_value('foo')
        'foo'

        >>> abjad.fsv.format_scheme_value('bar baz')
        '"bar baz"'

        >>> abjad.fsv.format_scheme_value([1.5, True, False])
        '(1.5 #t #f)'

    ..  container:: example

        Strings without whitespace can be forcibly quoted via the
        ``force_quotes`` keyword:

        >>> abjad.fsv.format_scheme_value(
        ...     'foo',
        ...     force_quotes=True,
        ...     )
        '"foo"'

    ..  container:: example

        Set verbatim to true to format value exactly (with only hash
        preprended):

        >>> string = '(lambda (grob) (grob-interpret-markup grob'
        >>> string += r' #{ \markup \musicglyph #"noteheads.s0harmonic" #}))'
        >>> abjad.fsv.format_scheme_value(string, verbatim=True)
        '(lambda (grob) (grob-interpret-markup grob #{ \\markup \\musicglyph #"noteheads.s0harmonic" #}))'

    ..  container:: example

        Hash symbol at the beginning of a string does not result in quoted
        output:

        >>> string = '#1-finger'
        >>> abjad.fsv.format_scheme_value(string)
        '#1-finger'

    """
    if isinstance(value, str) and verbatim:
        return value
    elif isinstance(value, str) and not verbatim:
        value = value.replace('"', r"\"")
        if value.startswith("#"):
            pass
        elif value.startswith("\\"):
            pass
        elif force_quotes or " " in value or "#" in value:
            return f'"{value}"'
        return value
    elif value is True:
        return "#t"
    elif value is False:
        return "#f"
    elif isinstance(value, (list, tuple)):
        string = " ".join(format_scheme_value(_) for _ in value)
        return f"({string})"
    elif value is None:
        return "#f"
    return str(value)



from .. import enums, math
from ..bundle import LilyPondFormatBundle
from ..storage import FormatSpecification, StorageFormatManager


class StemTremolo:
    r"""
    Stem tremolo.

    ..  container:: example

        Sixteenth-note tremolo:

        >>> note = abjad.Note("c'4")
        >>> stem_tremolo = abjad.StemTremolo(16)
        >>> abjad.attach(stem_tremolo, note)
        >>> abjad.show(note) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(note)
            >>> print(string)
            c'4
            :16

    ..  container:: example

        Thirty-second-note tremolo:

        >>> note = abjad.Note("c'4")
        >>> stem_tremolo = abjad.StemTremolo(32)
        >>> abjad.attach(stem_tremolo, note)
        >>> abjad.show(note) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(note)
            >>> print(string)
            c'4
            :32

    ..  container:: example

        REGRESSION. Consider a note, rest, chord to which a stem tremolo is attached.
        When such a note, rest, chord splits into two notes, rests, chords then a stem
        tremolo attaches to each of the resultant notes, rests, chords:

        >>> staff = abjad.Staff("c'4 c'2.")
        >>> abjad.attach(abjad.StemTremolo(), staff[1])
        >>> abjad.show(staff) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(staff)
            >>> print(string)
            \new Staff
            {
                c'4
                c'2.
                :16
            }

        >>> abjad.Meter.rewrite_meter(staff[:], abjad.Meter((3, 4)))
        >>> abjad.show(staff) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(staff)
            >>> print(string)
            \new Staff
            {
                c'4
                c'2
                :16
                ~
                c'4
                :16
            }

    """

    ### CLASS VARIABLES ###

    __slots__ = ("_tremolo_flags",)

    _format_slot = "after"

    _time_orientation = enums.Middle

    ### INITIALIZER ###

    def __init__(self, tremolo_flags: int = 16) -> None:
        if isinstance(tremolo_flags, type(self)):
            tremolo_flags = tremolo_flags.tremolo_flags
        tremolo_flags = int(tremolo_flags)
        if not math.is_nonnegative_integer_power_of_two(tremolo_flags):
            raise ValueError(f"nonnegative integer power of 2: {tremolo_flags!r}.")
        self._tremolo_flags = tremolo_flags

    ### SPECIAL METHODS ###

    def __eq__(self, argument) -> bool:
        """
        Is true when all initialization values of Abjad value object equal
        the initialization values of ``argument``.
        """
        return StorageFormatManager.compare_objects(self, argument)

    def __hash__(self) -> int:
        """
        Hashes Abjad value object.
        """
        hash_values = StorageFormatManager(self).get_hash_values()
        try:
            result = hash(hash_values)
        except TypeError:
            raise TypeError(f"unhashable type: {self}")
        return result

    def __repr__(self) -> str:
        """
        Gets interpreter representation.
        """
        return StorageFormatManager(self).get_repr_format()

    def __str__(self) -> str:
        """
        Gets string representation of stem tremolo.

        ..  container:: example

            Sixteenth-note tremolo:

            >>> stem_tremolo = abjad.StemTremolo(16)
            >>> print(str(stem_tremolo))
            :16

        ..  container:: example

            Thirty-second-note tremolo:

            >>> stem_tremolo = abjad.StemTremolo(32)
            >>> print(str(stem_tremolo))
            :32

        """
        return f":{self.tremolo_flags!s}"

    ### PRIVATE METHODS ###

    def _get_format_specification(self):
        return FormatSpecification(client=self, storage_format_is_indented=False)

    def _get_lilypond_format(self):
        return str(self)

    def _get_lilypond_format_bundle(self, component=None):
        bundle = LilyPondFormatBundle()
        bundle.after.stem_tremolos.append(self._get_lilypond_format())
        return bundle

    ### PUBLIC PROPERTIES ###

    @property
    def tremolo_flags(self) -> int:
        """
        Gets tremolo flags of stem tremolo.

        ..  container:: example

            Sixteenth-note tremolo:

            >>> stem_tremolo = abjad.StemTremolo(16)
            >>> stem_tremolo.tremolo_flags
            16

        ..  container:: example

            Thirty-second-note tremolo:

            >>> stem_tremolo = abjad.StemTremolo(32)
            >>> stem_tremolo.tremolo_flags
            32

        Set to nonnegative integer power of 2.
        """
        return self._tremolo_flags

    @property
    def tweaks(self) -> None:
        r"""
        Are not implemented on stem tremolo.

        The LilyPond ``:`` command refuses tweaks.

        Override the LilyPond ``StemTremolo`` grob instead.
        """
        pass



import collections
import copy
import importlib
import inspect
import numbers
import os
import pathlib
import subprocess
import time
import typing

from . import _inspect
from . import tag as _tag
from .attach import attach
from .bundle import LilyPondFormatBundle
from .configuration import Configuration
from .contextmanagers import TemporaryDirectoryChange
from .indicators.LilyPondComment import LilyPondComment
from .indicators.TimeSignature import TimeSignature
from .iterate import Iteration
from .lilypond import lilypond
from .markups import Markup
from .overrides import LilyPondLiteral, override, setting
from .pitch.pitches import NamedPitch
from .score import Component, Container, Context, Leaf, Score, Skip, Staff, Voice
from .select import Selection
from .sequence import Sequence
from .storage import FormatSpecification, StorageFormatManager

configuration = Configuration()


class Block:
    r"""
    LilyPond file block.

    ..  container:: example

        REGRESSION. Blocks remember attribute assignment order.

        Here right margin precedes left margin even though left margin
        alphabetizes before right margin:

        >>> block = abjad.Block(name="paper")
        >>> block.right_margin = abjad.LilyPondDimension(2, "cm")
        >>> block.left_margin = abjad.LilyPondDimension(2, "cm")
        >>> block
        <Block(name='paper')>

        >>> string = abjad.lilypond(block)
        >>> print(string)
        \paper {
            right-margin = 2\cm
            left-margin = 2\cm
        }

    ..  container:: example

        >>> block = abjad.Block(name='score')
        >>> markup = abjad.Markup('foo')
        >>> block.items.append(markup)
        >>> block
        <Block(name='score')>

        >>> string = abjad.lilypond(block)
        >>> print(string)
        \score {
            {
                \markup { foo }
            }
        }

    """

    ### INITIALIZER ###

    def __init__(self, name="score"):
        assert isinstance(name, str), repr(name)
        self._name = name
        escaped_name = rf"\{name}"
        self._escaped_name = escaped_name
        self._items = []
        self._public_attribute_names = []

    ### SPECIAL METHODS ###

    def __delattr__(self, name) -> None:
        """
        Deletes block attribute with ``name``.

        ..  container:: example

            >>> header_block = abjad.Block(name="header")
            >>> header_block.tagline = False
            >>> header_block.tagline
            False

            >>> delattr(header_block, "tagline")
            >>> hasattr(header_block, "tagline")
            False

        """
        self._public_attribute_names.remove(name)
        object.__delattr__(self, name)

    def __getitem__(self, name):
        """
        Gets item with ``name``.

        ..  container:: example

            Gets score with name ``'Red Example Score'`` in score block:

            >>> block = abjad.Block(name="score")
            >>> score = abjad.Score(name="Red_Example_Score")
            >>> block.items.append(score)

            >>> block["Red_Example_Score"]
            Score(simultaneous=True, name='Red_Example_Score')

        Returns item.

        Raises key error when no item with ``name`` is found.
        """
        for item in self.items:
            if getattr(item, "name", None) == name:
                return item
        raise KeyError

    def __repr__(self) -> str:
        """
        Gets interpreter representation.
        """
        return StorageFormatManager(self).get_repr_format()

    def __setattr__(self, name, value):
        """
        Sets block ``name`` to ``value``.

        Returns none.
        """
        if not name.startswith("_") and name not in self._public_attribute_names:
            self._public_attribute_names.append(name)
        object.__setattr__(self, name, value)

    def __setstate__(self, state):
        """
        Sets state of block.

        Returns none.
        """
        if not hasattr(self, "_public_attribute_names"):
            self._public_attribute_names = []
        for key, value in state.items():
            setattr(self, key, value)

    ### PRIVATE METHODS ###

    def _format_item(self, item, depth=1):
        indent = LilyPondFormatBundle.indent * depth
        result = []
        if isinstance(item, (list, tuple)):
            result.append(indent + "{")
            depth_ = depth + 1
            for x in item:
                pieces = self._format_item(x, depth=depth_)
                result.extend(pieces)
            result.append(indent + "}")
        elif isinstance(item, str):
            if item.isspace():
                string = ""
            else:
                string = indent + item
            result.append(string)
        elif "_get_format_pieces" in dir(item):
            try:
                pieces = item._get_format_pieces()
            except TypeError:
                pieces = item._get_format_pieces()
            for piece in pieces:
                if piece.isspace():
                    piece = ""
                else:
                    piece = indent + piece
                result.append(piece)
        return result

    def _formatted_context_blocks(self):
        result = []
        context_blocks = []
        for item in self.items:
            if isinstance(item, ContextBlock):
                context_blocks.append(item)
        for context_block in context_blocks:
            result.extend(context_block._get_format_pieces())
        return result

    def _get_format_pieces(self, tag=None):
        indent = LilyPondFormatBundle.indent
        result = []
        if (
            not self._get_formatted_user_attributes()
            and not getattr(self, "contexts", None)
            and not getattr(self, "context_blocks", None)
            and not len(self.items)
        ):
            string = f"{self._escaped_name} {{}}"
            result.append(string)
            return result
        strings = [f"{self._escaped_name}", "{"]
        if tag is not None:
            strings = _tag.double_tag(strings, tag)
        result.extend(strings)
        for item in self.items:
            if isinstance(item, ContextBlock):
                continue
            if isinstance(item, (Leaf, Markup)):
                item = [item]
            result.extend(self._format_item(item))
        formatted_attributes = self._get_formatted_user_attributes()
        formatted_attributes = [indent + _ for _ in formatted_attributes]
        result.extend(formatted_attributes)
        formatted_context_blocks = self._formatted_context_blocks()
        formatted_context_blocks = [indent + _ for _ in formatted_context_blocks]
        result.extend(formatted_context_blocks)
        string = "}"
        strings = [string]
        if tag is not None:
            strings = _tag.double_tag(strings, tag)
        result.extend(strings)

        return result

    def _get_format_specification(self):
        return FormatSpecification(
            client=self, repr_is_bracketed=True, repr_is_indented=False
        )

    def _get_formatted_user_attributes(self):
        result = []
        for key in self._public_attribute_names:
            assert not key.startswith("_"), repr(key)
            value = getattr(self, key)
            formatted_key = key.split("__")
            for i, k in enumerate(formatted_key):
                formatted_key[i] = k.replace("_", "-")
                if 0 < i:
                    string = f"#'{formatted_key[i]}"
                    formatted_key[i] = string
            formatted_key = " ".join(formatted_key)
            if isinstance(value, Markup):
                formatted_value = value._get_format_pieces()
            elif isinstance(value, LilyPondDimension):
                formatted_value = [lilypond(value)]
            else:
                assert isinstance(value, (int, str, float)), repr((self, key, value))
                formatted_value = [str(value)]
            setting = f"{formatted_key!s} = {formatted_value[0]!s}"
            result.append(setting)
            result.extend(formatted_value[1:])
        return result

    def _get_lilypond_format(self, tag=None):
        return "\n".join(self._get_format_pieces(tag=tag))

    ### PUBLIC PROPERTIES ###

    @property
    def items(self):
        r"""
        Gets items in block.

        ..  container:: example

            >>> block = abjad.Block(name="score")
            >>> markup = abjad.Markup("foo")
            >>> block.items.append(markup)

            >>> block.items
            [Markup(contents=['foo'])]

        ..  container:: example

            Accepts strings:

            >>> staff = abjad.Staff("c'4 d' e' f'")
            >>> score_block = abjad.Block(name="score")
            >>> score_block.items.append("<<")
            >>> score_block.items.append(r'{ \include "layout.ly" }')
            >>> score_block.items.append(staff)
            >>> score_block.items.append(">>")
            >>> lilypond_file = abjad.LilyPondFile(
            ...     lilypond_language_token=False,
            ...     lilypond_version_token=False,
            ... )
            >>> lilypond_file.items.append(score_block)

            >>> string = abjad.lilypond(lilypond_file)
            >>> print(string)
            \score {
                <<
                { \include "layout.ly" }
                \new Staff
                {
                    c'4
                    d'4
                    e'4
                    f'4
                }
                >>
            }

        Returns list.
        """
        return self._items

    @property
    def name(self) -> typing.Optional[str]:
        """
        Gets name of block.

        ..  container:: example

            >>> block = abjad.Block(name="score")
            >>> markup = abjad.Markup("foo")
            >>> block.items.append(markup)

            >>> block.name
            'score'

        """
        return self._name

    ### PUBLIC METHODS ###

    def empty(self) -> bool:
        """
        Is true when block contains no items and has no user attributes.
        """
        if not self.items and not self._get_formatted_user_attributes():
            return True
        return False


class ContextBlock(Block):
    r"""
    LilyPond file ``\context`` block.

    ..  container:: example

        >>> block = abjad.ContextBlock(
        ...     source_lilypond_type="Staff",
        ...     name="FluteStaff",
        ...     type_="Engraver_group",
        ...     alias="Staff",
        ... )
        >>> block.remove_commands.append("Forbid_line_break_engraver")
        >>> block.consists_commands.append("Horizontal_bracket_engraver")
        >>> block.accepts_commands.append("FluteUpperVoice")
        >>> block.accepts_commands.append("FluteLowerVoice")
        >>> block.items.append(r"\accidentalStyle dodecaphonic")
        >>> abjad.override(block).Beam.positions = "#'(-4 . -4)"
        >>> abjad.override(block).Stem.stem_end_position = -6
        >>> abjad.setting(block).autoBeaming = False
        >>> abjad.setting(block).tupletFullLength = True
        >>> block
        <ContextBlock(source_lilypond_type='Staff', name='FluteStaff', type_='Engraver_group', alias='Staff')>

        >>> print(abjad.lilypond(block))
        \context {
            \Staff
            \name FluteStaff
            \type Engraver_group
            \alias Staff
            \remove Forbid_line_break_engraver
            \consists Horizontal_bracket_engraver
            \accepts FluteUpperVoice
            \accepts FluteLowerVoice
            \override Beam.positions = #'(-4 . -4)
            \override Stem.stem-end-position = -6
            autoBeaming = ##f
            tupletFullLength = ##t
            \accidentalStyle dodecaphonic
        }

    """

    ### INITIALIZER ###

    def __init__(self, source_lilypond_type=None, name=None, type_=None, alias=None):
        Block.__init__(self, name="context")
        self._source_lilypond_type = source_lilypond_type
        self._name = name
        self._type_ = type_
        self._alias = alias
        self._accepts_commands = []
        self._consists_commands = []
        self._remove_commands = []

    ### PRIVATE METHODS ###

    def _get_format_pieces(self, tag=None):
        indent = LilyPondFormatBundle.indent
        result = []
        string = f"{self._escaped_name} {{"
        result.append(string)
        # CAUTION: source context name must come before type_ to allow
        # context redefinition.
        if self.source_lilypond_type is not None:
            string = indent + rf"\{self.source_lilypond_type}"
            result.append(string)
        if self.name is not None:
            string = indent + rf"\name {self.name}"
            result.append(string)
        if self.type_ is not None:
            string = indent + rf"\type {self.type_}"
            result.append(string)
        if self.alias is not None:
            string = indent + rf"\alias {self.alias}"
            result.append(string)
        for statement in self.remove_commands:
            string = indent + rf"\remove {statement}"
            result.append(string)
        # CAUTION: LilyPond \consists statements are order-significant!
        for statement in self.consists_commands:
            string = indent + rf"\consists {statement}"
            result.append(string)
        for statement in self.accepts_commands:
            string = indent + rf"\accepts {statement}"
            result.append(string)
        overrides = override(self)._list_format_contributions("override")
        for statement in overrides:
            string = indent + statement
            result.append(string)
        setting_contributions = setting(self)._format_in_with_block()
        for setting_contribution in sorted(setting_contributions):
            string = indent + setting_contribution
            result.append(string)
        for item in self.items:
            if isinstance(item, str):
                string = indent + f"{item}"
                result.append(string)
            elif "_get_format_pieces" in dir(item):
                for piece in item._get_format_pieces():
                    if piece.isspace():
                        piece = ""
                    else:
                        piece = indent + piece
                    result.append(piece)
            else:
                pass
        result.append("}")
        return result

    ### PUBLIC PROPERTIES ###

    @property
    def accepts_commands(self) -> typing.List[str]:
        r"""
        Gets arguments of LilyPond ``\accepts`` commands.

        ..  container:: example

            >>> block = abjad.ContextBlock(
            ...     source_lilypond_type="Staff",
            ...     name="FluteStaff",
            ...     type_="Engraver_group",
            ...     alias="Staff",
            ... )
            >>> block.remove_commands.append("Forbid_line_break_engraver")
            >>> block.consists_commands.append("Horizontal_bracket_engraver")
            >>> block.accepts_commands.append("FluteUpperVoice")
            >>> block.accepts_commands.append("FluteLowerVoice")
            >>> block.items.append(r"\accidentalStyle dodecaphonic")
            >>> abjad.override(block).Beam.positions = "#'(-4 . -4)"
            >>> abjad.override(block).Stem.stem_end_position = -6
            >>> abjad.setting(block).autoBeaming = False
            >>> abjad.setting(block).tupletFullLength = True

            >>> block.accepts_commands
            ['FluteUpperVoice', 'FluteLowerVoice']

        """
        return self._accepts_commands

    @property
    def alias(self) -> typing.Optional[str]:
        r"""
        Gets and sets argument of LilyPond ``\alias`` command.

        ..  container:: example

            >>> block = abjad.ContextBlock(
            ...     source_lilypond_type="Staff",
            ...     name="FluteStaff",
            ...     type_="Engraver_group",
            ...     alias="Staff",
            ... )
            >>> block.remove_commands.append("Forbid_line_break_engraver")
            >>> block.consists_commands.append("Horizontal_bracket_engraver")
            >>> block.accepts_commands.append("FluteUpperVoice")
            >>> block.accepts_commands.append("FluteLowerVoice")
            >>> block.items.append(r"\accidentalStyle dodecaphonic")
            >>> abjad.override(block).Beam.positions = "#'(-4 . -4)"
            >>> abjad.override(block).Stem.stem_end_position = -6
            >>> abjad.setting(block).autoBeaming = False
            >>> abjad.setting(block).tupletFullLength = True

            >>> block.alias
            'Staff'

        """
        return self._alias

    @property
    def consists_commands(self) -> typing.List[str]:
        r"""
        Gets arguments of LilyPond ``\consists`` commands.

        ..  container:: example

            >>> block = abjad.ContextBlock(
            ...     source_lilypond_type="Staff",
            ...     name="FluteStaff",
            ...     type_="Engraver_group",
            ...     alias="Staff",
            ... )
            >>> block.remove_commands.append("Forbid_line_break_engraver")
            >>> block.consists_commands.append("Horizontal_bracket_engraver")
            >>> block.accepts_commands.append("FluteUpperVoice")
            >>> block.accepts_commands.append("FluteLowerVoice")
            >>> block.items.append(r"\accidentalStyle dodecaphonic")
            >>> abjad.override(block).Beam.positions = "#'(-4 . -4)"
            >>> abjad.override(block).Stem.stem_end_position = -6
            >>> abjad.setting(block).autoBeaming = False
            >>> abjad.setting(block).tupletFullLength = True

            >>> block.consists_commands
            ['Horizontal_bracket_engraver']

        """
        return self._consists_commands

    @property
    def items(self) -> typing.List[str]:
        r"""
        Gets items in context block.

        ..  container:: example

            >>> block = abjad.ContextBlock(
            ...     source_lilypond_type="Staff",
            ...     name="FluteStaff",
            ...     type_="Engraver_group",
            ...     alias="Staff",
            ... )
            >>> block.remove_commands.append("Forbid_line_break_engraver")
            >>> block.consists_commands.append("Horizontal_bracket_engraver")
            >>> block.accepts_commands.append("FluteUpperVoice")
            >>> block.accepts_commands.append("FluteLowerVoice")
            >>> block.items.append(r"\accidentalStyle dodecaphonic")
            >>> abjad.override(block).Beam.positions = "#'(-4 . -4)"
            >>> abjad.override(block).Stem.stem_end_position = -6
            >>> abjad.setting(block).autoBeaming = False
            >>> abjad.setting(block).tupletFullLength = True

            >>> block.items
            ['\\accidentalStyle dodecaphonic']

        """
        return self._items

    @property
    def name(self) -> typing.Optional[str]:
        r"""
        Gets and sets argument of LilyPond ``\name`` command.

        ..  container:: example

            >>> block = abjad.ContextBlock(
            ...     source_lilypond_type="Staff",
            ...     name="FluteStaff",
            ...     type_="Engraver_group",
            ...     alias="Staff",
            ... )
            >>> block.remove_commands.append("Forbid_line_break_engraver")
            >>> block.consists_commands.append("Horizontal_bracket_engraver")
            >>> block.accepts_commands.append("FluteUpperVoice")
            >>> block.accepts_commands.append("FluteLowerVoice")
            >>> block.items.append(r"\accidentalStyle dodecaphonic")
            >>> abjad.override(block).Beam.positions = "#'(-4 . -4)"
            >>> abjad.override(block).Stem.stem_end_position = -6
            >>> abjad.setting(block).autoBeaming = False
            >>> abjad.setting(block).tupletFullLength = True

            >>> block.name
            'FluteStaff'

        """
        return self._name

    @property
    def remove_commands(self) -> typing.List[str]:
        r"""
        Gets arguments of LilyPond ``\remove`` commands.

        ..  container:: example

            >>> block = abjad.ContextBlock(
            ...     source_lilypond_type="Staff",
            ...     name="FluteStaff",
            ...     type_="Engraver_group",
            ...     alias="Staff",
            ... )
            >>> block.remove_commands.append("Forbid_line_break_engraver")
            >>> block.consists_commands.append("Horizontal_bracket_engraver")
            >>> block.accepts_commands.append("FluteUpperVoice")
            >>> block.accepts_commands.append("FluteLowerVoice")
            >>> block.items.append(r"\accidentalStyle dodecaphonic")
            >>> abjad.override(block).Beam.positions = "#'(-4 . -4)"
            >>> abjad.override(block).Stem.stem_end_position = -6
            >>> abjad.setting(block).autoBeaming = False
            >>> abjad.setting(block).tupletFullLength = True

            >>> block.remove_commands
            ['Forbid_line_break_engraver']

        """
        return self._remove_commands

    @property
    def source_lilypond_type(self) -> typing.Optional[str]:
        r"""
        Gets and sets source context name.

        ..  container:: example

            >>> block = abjad.ContextBlock(
            ...     source_lilypond_type="Staff",
            ...     name="FluteStaff",
            ...     type_="Engraver_group",
            ...     alias="Staff",
            ... )
            >>> block.remove_commands.append("Forbid_line_break_engraver")
            >>> block.consists_commands.append("Horizontal_bracket_engraver")
            >>> block.accepts_commands.append("FluteUpperVoice")
            >>> block.accepts_commands.append("FluteLowerVoice")
            >>> block.items.append(r"\accidentalStyle dodecaphonic")
            >>> abjad.override(block).Beam.positions = "#'(-4 . -4)"
            >>> abjad.override(block).Stem.stem_end_position = -6
            >>> abjad.setting(block).autoBeaming = False
            >>> abjad.setting(block).tupletFullLength = True

            >>> block.source_lilypond_type
            'Staff'

        """
        return self._source_lilypond_type

    @property
    def type_(self) -> typing.Optional[str]:
        r"""
        Gets and sets argument of LilyPond ``\type`` command.

        ..  container:: example

            >>> block = abjad.ContextBlock(
            ...     source_lilypond_type="Staff",
            ...     name="FluteStaff",
            ...     type_="Engraver_group",
            ...     alias="Staff",
            ... )
            >>> block.remove_commands.append("Forbid_line_break_engraver")
            >>> block.consists_commands.append("Horizontal_bracket_engraver")
            >>> block.accepts_commands.append("FluteUpperVoice")
            >>> block.accepts_commands.append("FluteLowerVoice")
            >>> block.items.append(r"\accidentalStyle dodecaphonic")
            >>> abjad.override(block).Beam.positions = "#'(-4 . -4)"
            >>> abjad.override(block).Stem.stem_end_position = -6
            >>> abjad.setting(block).autoBeaming = False
            >>> abjad.setting(block).tupletFullLength = True

            >>> block.type_
            'Engraver_group'

        """
        return self._type_


class DateTimeToken:
    """
    LilyPond file date / time token.

    ..  container:: example

        >>> abjad.DateTimeToken()
        DateTimeToken()

    """

    ### CLASS VARIABLES ###

    __slots__ = ("_date_string",)

    ### INITIALIZER ###

    def __init__(self, date_string=None):
        assert isinstance(date_string, (str, type(None)))
        self._date_string = date_string

    ### SPECIAL METHODS ###

    def __repr__(self) -> str:
        """
        Gets interpreter representation of date / time token.

        ..  container:: example

            >>> abjad.DateTimeToken()
            DateTimeToken()

        """
        date_string = self._date_string or ""
        return f"{type(self).__name__}({date_string})"

    ### PRIVATE METHODS ###

    def _get_lilypond_format(self):
        return self.date_string

    ### PUBLIC PROPERTIES ###

    @property
    def date_string(self) -> str:
        """
        Gets date string of date / time token.

        ..  container:: example

            >>> token = abjad.DateTimeToken()
            >>> token.date_string # doctest: +SKIP
            '2014-01-23 12:21'

        """
        date_string = self._date_string or time.strftime("%Y-%m-%d %H:%M")
        return date_string


class LilyPondDimension:
    r"""
    LilyPond file ``\paper`` block dimension.

    ..  container:: example

        >>> abjad.LilyPondDimension(2, "in")
        LilyPondDimension(value=2, unit='in')

    Use for LilyPond file ``\paper`` block attributes.
    """

    ### CLASS VARIABLES ###

    __slots__ = ("_unit", "_value")

    ### INITIALIZER ###

    def __init__(self, value=0, unit="cm"):
        assert isinstance(value, numbers.Number) and 0 <= value
        assert unit in ("cm", "in", "mm", "pt")
        self._value = value
        self._unit = unit

    ### SPECIAL METHODS ###

    def __repr__(self) -> str:
        """
        Gets interpreter representation.
        """
        return StorageFormatManager(self).get_repr_format()

    ### PRIVATE METHODS ###

    def _get_format_pieces(self, tag=None):
        return [rf"{self.value}\{self.unit}"]

    def _get_lilypond_format(self):
        return "\n".join(self._get_format_pieces())

    ### PUBLIC PROPERTIES ###

    @property
    def unit(self) -> str:
        """
        Gets unit of LilyPond dimension.

        ..  container:: example

            >>> dimension = abjad.LilyPondDimension(2, "in")
            >>> dimension.unit
            'in'

        Returns ``'cm'``, ``'in'``, ``'mm'`` or ``'pt'``.
        """
        return self._unit

    @property
    def value(self) -> typing.Union[int, float]:
        """
        Gets value of LilyPond dimension.

        ..  container:: example

            >>> dimension = abjad.LilyPondDimension(2, "in")
            >>> dimension.value
            2

        """
        return self._value


class LilyPondLanguageToken:
    r"""
    LilyPond file ``\language`` token.

    ..  container:: example

        >>> abjad.LilyPondLanguageToken()
        LilyPondLanguageToken()

    """

    ### CLASS VARIABLES ###

    __slots__ = ()

    ### SPECIAL METHODS ###

    def __repr__(self) -> str:
        """
        Gets interpreter representation of LilyPond language token.

        ..  container:: example

            >>> token = abjad.LilyPondLanguageToken()
            >>> token
            LilyPondLanguageToken()

        """
        return f"{type(self).__name__}()"

    ### PRIVATE METHODS ###

    def _get_lilypond_format(self):
        string = r'\language "english"'
        return string


class LilyPondVersionToken:
    r"""
    LilyPond file ``\version`` token.
    """

    ### CLASS VARIABLES ###

    __slots__ = ("_version_string",)

    ### INITIALIZER ###

    def __init__(self, version_string=None):
        assert isinstance(version_string, (str, type(None)))
        if version_string is None:
            version_string = configuration.get_lilypond_version_string()
        self._version_string = version_string

    ### SPECIAL METHODS ###

    def __repr__(self) -> str:
        """
        Gets interpreter representation of LilyPond version_string token.

        ..  container:: example

            >>> token = abjad.LilyPondVersionToken()
            >>> token # doctest: +SKIP
            LilyPondVersionToken('2.19.84')

        """
        return f"{type(self).__name__}({self.version_string!r})"

    ### PRIVATE METHODS ###

    def _get_lilypond_format(self):
        return rf'\version "{self.version_string}"'

    ### PUBLIC PROPERTIES ###

    @property
    def version_string(self) -> str:
        """
        Gets version string of LilyPond version token.

        ..  container:: example

            Gets version string from install environment:

            >>> token = abjad.LilyPondVersionToken(
            ...     version_string=None,
            ...     )
            >>> token.version_string # doctest: +SKIP
            '2.19.84'

        ..  container:: example

            Gets version string from explicit input:

            >>> token = abjad.LilyPondVersionToken(
            ...     version_string="2.19.84",
            ...     )
            >>> token.version_string
            '2.19.84'

        """
        return self._version_string


class PackageGitCommitToken:
    """
    Python package git commit token.

    ..  container:: example

        >>> token = abjad.PackageGitCommitToken("abjad")
        >>> token
        PackageGitCommitToken(package_name='abjad')

        >>> print(abjad.lilypond(token))  # doctest: +SKIP
        package "abjad" @ b6a48a7 [implement-lpf-git-token] (2016-02-02 13:36:25)

    """

    ### CLASS VARIABLES ###

    __slots__ = ("_package_name",)

    ### INITIALIZER ###

    def __init__(self, package_name=None):
        self._package_name = package_name

    ### SPECIAL METHODS ###

    def __repr__(self) -> str:
        """
        Gets interpreter representation.
        """
        return StorageFormatManager(self).get_repr_format()

    ### PRIVATE METHODS ###

    def _get_commit_timestamp(self, commit_hash):
        command = f"git show -s --format=%ci {commit_hash}"
        return self._run_command(command)

    def _get_git_branch(self):
        command = "git rev-parse --abbrev-ref HEAD"
        return self._run_command(command)

    def _get_git_hash(self):
        command = "git rev-parse HEAD"
        return self._run_command(command)

    def _get_lilypond_format(self):
        path = self._get_package_path()
        with TemporaryDirectoryChange(path):
            git_branch = self._get_git_branch()
            git_hash = self._get_git_hash()
            timestamp = self._get_commit_timestamp(git_hash)
        date, time, _ = timestamp.split()
        return 'package "{}" @ {} [{}] ({} {})'.format(
            self._package_name, git_hash[:7], git_branch, date, time
        )

    def _get_package_path(self):
        module = importlib.import_module(self._package_name)
        path = module.__path__[0]
        if not os.path.isdir(path):
            path = os.path.dirname(path)
        path = os.path.abspath(path)
        return path

    def _run_command(self, command):
        process = subprocess.Popen(
            command, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE
        )
        process.wait()
        if process.returncode:
            return None
        result = process.stdout.read().splitlines()[0]
        result = result.decode("utf-8")
        return result

    ### PUBLIC PROPERTIES ###

    @property
    def package_name(self) -> str:
        """
        Gets package name of package git commit token.

        ..  container:: example

            >>> token = abjad.PackageGitCommitToken("abjad")
            >>> token.package_name
            'abjad'

        """
        return self._package_name


class LilyPondFile:
    r"""
    LilyPond file.

    ..  container:: example

        >>> staff = abjad.Staff("c'8 d'8 e'8 f'8")
        >>> comments = [
        ...     "File construct as an example.",
        ...     "Parts shown here for positioning.",
        ... ]
        >>> includes = [
        ...     "external-settings-file-1.ily",
        ...     "external-settings-file-2.ily",
        ... ]
        >>> lilypond_file = abjad.LilyPondFile(
        ...     items=[staff],
        ...     default_paper_size=("a5", "portrait"),
        ...     comments=comments,
        ...     includes=includes,
        ...     global_staff_size=16,
        ... )

        >>> abjad.show(lilypond_file) # doctest: +SKIP

        ::

            >>> print(abjad.lilypond(lilypond_file)) # doctest: +SKIP
            % File construct as an example.
            % Parts shown here for positioning.
            <BLANKLINE>
            \version "2.23.0"
            \language "english"
            <BLANKLINE>
            \include "external-settings-file-1.ily"
            \include "external-settings-file-2.ily"
            <BLANKLINE>
            #(set-default-paper-size "a5" 'portrait)
            #(set-global-staff-size 16)
            <BLANKLINE>
            \new Staff
            {
                c'8
                d'8
                e'8
                f'8
            }

    """

    ### CLASS VARIABLES ###

    __slots__ = (
        "_comments",
        "_date_time_token",
        "_default_paper_size",
        "_global_staff_size",
        "_includes",
        "_items",
        "_lilypond_language_token",
        "_lilypond_version_token",
        "_tag",
        "_use_relative_includes",
    )

    ### INITIALIZER ###

    def __init__(
        self,
        comments=None,
        date_time_token=None,
        default_paper_size=None,
        global_staff_size=None,
        includes=None,
        items=None,
        lilypond_language_token=None,
        lilypond_version_token=None,
        tag: _tag.Tag = None,
        use_relative_includes=None,
    ) -> None:
        comments = comments or ()
        comments = tuple(comments)
        self._comments = comments
        self._date_time_token = None
        if bool(date_time_token):
            self._date_time_token = DateTimeToken()
        self._default_paper_size = default_paper_size
        self._global_staff_size = global_staff_size
        includes = list(includes or [])
        self._includes = includes
        self._items = list(items or [])
        self._lilypond_language_token = None
        if lilypond_language_token is not False:
            language = LilyPondLanguageToken()
            self._lilypond_language_token = language
        self._lilypond_version_token = None
        if lilypond_version_token is not False:
            version = LilyPondVersionToken()
            self._lilypond_version_token = version
        if tag is not None:
            assert isinstance(tag, _tag.Tag), repr(tag)
        self._tag = tag
        self._use_relative_includes = use_relative_includes

    ### SPECIAL METHODS ###

    def __contains__(self, argument) -> bool:
        """
        Is true when LilyPond file contains ``argument``.

        ..  container:: example

            >>> staff = abjad.Staff("c'4 d' e' f'")
            >>> lilypond_file = abjad.LilyPondFile(items=[staff])

            >>> staff in lilypond_file
            True

            >>> abjad.Staff in lilypond_file
            True

            >>> "Allegro" in lilypond_file
            False

            >>> 0 in lilypond_file
            False

        """
        try:
            self[argument]
            return True
        except (AssertionError, KeyError, ValueError, TypeError):
            return False

    def __getitem__(self, name):
        r"""
        Gets item with ``name``.

        ..  container:: example

            Searches score:

            >>> voice_1 = abjad.Voice("c''4 b' a' g'", name="Custom_Voice_1")
            >>> literal = abjad.LilyPondLiteral(r"\voiceOne", "opening")
            >>> abjad.attach(literal, voice_1)
            >>> voice_2 = abjad.Voice("c'4 d' e' f'", name="Custom_Voice_2")
            >>> literal = abjad.LilyPondLiteral(r"\voiceTwo", "opening")
            >>> abjad.attach(literal, voice_2)
            >>> staff = abjad.Staff(
            ...     [voice_1, voice_2],
            ...     simultaneous=True,
            ...     name="Custom_Staff",
            ... )
            >>> score = abjad.Score([staff], name="Custom_Score")
            >>> block = abjad.Block(name="score")
            >>> block.items.append(score)
            >>> lilypond_file = abjad.LilyPondFile(items=[block])
            >>> abjad.show(score) # doctest: +SKIP

            ..  docs::

                >>> string = abjad.lilypond(score)
                >>> print(string)
                \context Score = "Custom_Score"
                <<
                    \context Staff = "Custom_Staff"
                    <<
                        \context Voice = "Custom_Voice_1"
                        {
                            \voiceOne
                            c''4
                            b'4
                            a'4
                            g'4
                        }
                        \context Voice = "Custom_Voice_2"
                        {
                            \voiceTwo
                            c'4
                            d'4
                            e'4
                            f'4
                        }
                    >>
                >>

            >>> lilypond_file["score"]
            <Block(name='score')>

            >>> lilypond_file["Custom_Score"]
            <Score-"Custom_Score"<<1>>>

            >>> lilypond_file[abjad.Score]
            <Score-"Custom_Score"<<1>>>

            >>> lilypond_file["Custom_Staff"]
            <Staff-"Custom_Staff"<<2>>>

            >>> lilypond_file[abjad.Staff]
            <Staff-"Custom_Staff"<<2>>>

            >>> lilypond_file["Custom_Voice_1"]
            Voice("c''4 b'4 a'4 g'4", name='Custom_Voice_1')

            >>> lilypond_file["Custom_Voice_2"]
            Voice("c'4 d'4 e'4 f'4", name='Custom_Voice_2')

            >>> lilypond_file[abjad.Voice]
            Voice("c''4 b'4 a'4 g'4", name='Custom_Voice_1')

        ..  container:: example

            REGRESSION. Works when score block contains parallel container:

                >>> include_container = abjad.Container()
                >>> string = r'\include "layout.ly"'
                >>> literal = abjad.LilyPondLiteral(string, "opening")
                >>> abjad.attach(literal, include_container)
                >>> staff = abjad.Staff("c'4 d' e' f'", name="Custom_Staff")
                >>> container = abjad.Container(
                ...     [include_container, staff],
                ...     simultaneous=True,
                ... )
                >>> block = abjad.Block(name="score")
                >>> block.items.append(container)
                >>> lilypond_file = abjad.LilyPondFile(
                ...     items=[block],
                ...     lilypond_language_token=False,
                ...     lilypond_version_token=False,
                ... )
                >>> string = abjad.lilypond(lilypond_file)
                >>> print(string)
                \score {
                    <<
                        {
                            \include "layout.ly"
                        }
                        \context Staff = "Custom_Staff"
                        {
                            c'4
                            d'4
                            e'4
                            f'4
                        }
                    >>
                }

                >>> lilypond_file[abjad.Staff]
                Staff("c'4 d'4 e'4 f'4", name='Custom_Staff')

                >>> lilypond_file['Custom_Staff']
                Staff("c'4 d'4 e'4 f'4", name='Custom_Staff')

        Returns item.

        Raises key error when no item with ``name`` is found.
        """
        if not isinstance(name, str):
            if inspect.isclass(name):
                assert issubclass(name, Component), repr(name)
            else:
                assert isinstance(name, Component), repr(name)
        score = None
        if self.score_block and self.score_block.items:
            items = self.score_block.items
            for container in Iteration(items).components(Container):
                if isinstance(container, Context):
                    score = container
                    break
        if isinstance(name, str):
            for item in self.items:
                if getattr(item, "name", None) == name:
                    return item
            if score is not None:
                if score.name == name:
                    return score
                context = score[name]
                return context
            raise KeyError(f"can not find item with name {name!r}.")
        elif isinstance(name, Component):
            for item in self.items:
                if item is name:
                    return item
            if score is not None:
                if score is name:
                    return score
                prototype = Context
                for context in Iteration(score).components(prototype):
                    if context is name:
                        return context
            raise KeyError(f"can not find {name}.")
        elif inspect.isclass(name) and issubclass(name, Component):
            for item in self.items:
                if isinstance(item, name):
                    return item
            if score is not None:
                if isinstance(score, name):
                    return score
                prototype = Context
                for context in Iteration(score).components(prototype):
                    if isinstance(context, name):
                        return context
            raise KeyError(f"can not find item of class {name}.")
        else:
            raise TypeError(name)

    def __illustrate__(self) -> "LilyPondFile":
        """
        Illustrates LilyPond file.

        Returns LilyPond file unchanged.
        """
        return self

    def __repr__(self) -> str:
        """
        Gets interpreter representation of LilyPond file.
        """
        return StorageFormatManager(self).get_repr_format()

    ### PRIVATE METHODS ###

    def _get_format_pieces(self, tag=None):
        result = []
        if self.date_time_token is not None:
            string = f"% {self.date_time_token}"
            result.append(string)
        result.extend(self._get_formatted_comments())
        includes = []
        if self.lilypond_version_token is not None:
            string = f"{self.lilypond_version_token._get_lilypond_format()}"
            includes.append(string)
        if self.lilypond_language_token is not None:
            string = f"{self.lilypond_language_token._get_lilypond_format()}"
            includes.append(string)
        tag = _tag.Tag("abjad.LilyPondFile._get_format_pieces()")
        includes = _tag.double_tag(includes, self.get_tag(tag))
        includes = "\n".join(includes)
        if includes:
            result.append(includes)
        postincludes = []
        if self.use_relative_includes:
            string = "#(ly:set-option 'relative-includes #t)"
            postincludes.append(string)
        postincludes.extend(self._get_formatted_includes())
        postincludes.extend(self._get_formatted_scheme_settings())
        result.extend(postincludes)
        strings = self._get_formatted_blocks()
        if strings:
            result.append("")
        result.extend(strings)
        return result

    ### PRIVATE METHODS ###

    def _get_format_specification(self):
        return FormatSpecification(client=self)

    def _get_formatted_blocks(self):
        result = []
        tag = _tag.Tag("abjad.LilyPondFile._get_formatted_blocks()")
        tag = self.get_tag(tag)
        for item in self.items:
            if "_get_lilypond_format" in dir(item) and not isinstance(item, str):
                try:
                    string = item._get_lilypond_format(tag=tag)
                except TypeError:
                    string = item._get_lilypond_format()
                if string:
                    result.append(string)
            else:
                result.append(str(item))
        return result

    def _get_formatted_comments(self):
        result = []
        for comment in self.comments:
            if "_get_lilypond_format" in dir(comment) and not isinstance(comment, str):
                lilypond_format = comment._get_lilypond_format()
                if lilypond_format:
                    string = f"% {comment}"
                    result.append(string)
            else:
                string = f"% {comment!s}"
                result.append(string)
        if result:
            result = ["\n".join(result)]
        return result

    def _get_formatted_includes(self):
        result = []
        tag = _tag.Tag("abjad.LilyPondFile._get_formatted_includes()")
        tag = self.get_tag(tag)
        for include in self.includes:
            if isinstance(include, str):
                string = rf'\include "{include}"'
                result.append(string)
            elif isinstance(include, pathlib.Path):
                string = rf'\include "{include!s}"'
                result.append(string)
            elif isinstance(include, LilyPondLiteral):
                string = str(include.argument)
                result.append(string)
            else:
                result.append(include._get_lilypond_format())
        if result:
            result = _tag.double_tag(result, tag)
            result = ["\n".join(result)]
        return result

    def _get_formatted_scheme_settings(self):
        result = []
        tag = _tag.Tag("abjad.LilyPondFile._get_formatted_scheme_settings()")
        tag = self.get_tag(tag)
        default_paper_size = self.default_paper_size
        if default_paper_size is not None:
            dimension, orientation = default_paper_size
            string = f'#(set-default-paper-size "{dimension}" \'{orientation})'
            result.append(string)
        global_staff_size = self.global_staff_size
        if global_staff_size is not None:
            string = f"#(set-global-staff-size {global_staff_size})"
            result.append(string)
        if result:
            result = _tag.double_tag(result, tag)
            result = ["\n".join(result)]
        return result

    def _get_lilypond_format(self):
        strings = self._get_format_pieces()
        string = "\n".join(strings)
        lines = []
        for line in string.split("\n"):
            if line.isspace():
                lines.append("")
            else:
                lines.append(line)
        return "\n".join(lines)

    @staticmethod
    def _make_global_context_block(font_size=3, minimum_distance=10, padding=4):
        assert isinstance(font_size, (int, float))
        assert isinstance(padding, (int, float))
        block = ContextBlock(name="Global_Context", type_="Engraver_group")
        block.consists_commands.append("Axis_group_engraver")
        block.consists_commands.append("Time_signature_engraver")
        time_signature_grob = override(block).time_signature
        time_signature_grob.X_extent = (0, 0)
        time_signature_grob.X_offset = "#ly:self-alignment-interface::x-aligned-on-self"
        time_signature_grob.Y_extent = (0, 0)
        time_signature_grob.break_align_symbol = False
        time_signature_grob.break_visibility = "#end-of-line-invisible"
        time_signature_grob.font_size = font_size
        time_signature_grob.self_alignment_X = "#center"
        grob = override(block).vertical_axis_group
        string = "#'((basic-distance . 0) (minimum-distance . {minimum_distance}) (padding . {padding}) (stretchability . 0))"
        grob.default_staff_staff_spacing = string
        return block

    ### PUBLIC PROPERTIES ###

    @property
    def comments(self) -> typing.List[LilyPondComment]:
        """
        Gets comments of Lilypond file.

        ..  container:: example

            >>> lilypond_file = abjad.LilyPondFile()
            >>> lilypond_file.comments
            []

        """
        return list(self._comments)

    @property
    def date_time_token(self) -> typing.Optional[DateTimeToken]:
        """
        Gets date-time token.

        ..  container:: example

            >>> lilypond_file = abjad.LilyPondFile(date_time_token=True)
            >>> lilypond_file.date_time_token
            DateTimeToken()

        """
        return self._date_time_token

    @property
    def default_paper_size(self):
        """
        Gets default paper size of LilyPond file.

        ..  container:: example

            >>> lilypond_file = abjad.LilyPondFile()
            >>> lilypond_file.default_paper_size is None
            True

        Returns pair or none.
        """
        return self._default_paper_size

    @property
    def global_staff_size(self) -> typing.Union[int, float]:
        """
        Gets global staff size of LilyPond file.

        ..  container:: example

            >>> lilypond_file = abjad.LilyPondFile()
            >>> lilypond_file.global_staff_size is None
            True

        """
        return self._global_staff_size

    @property
    def header_block(self) -> typing.Optional[Block]:
        """
        Gets header block.

        ..  container:: example

            >>> block = abjad.Block(name="header")
            >>> lilypond_file = abjad.LilyPondFile(items=[block])
            >>> lilypond_file.header_block
            <Block(name='header')>

        """
        for item in self.items:
            if isinstance(item, Block):
                if item.name == "header":
                    return item
        return None

    @property
    def includes(self) -> typing.List[str]:
        """
        Gets includes of LilyPond file.

        ..  container:: example

            >>> lilypond_file = abjad.LilyPondFile()
            >>> lilypond_file.includes
            []

        """
        return self._includes

    @property
    def items(self) -> typing.List:
        r"""
        Gets items in LilyPond file.

        ..  container:: example

            >>> lilypond_file = abjad.LilyPondFile()
            >>> lilypond_file.items
            []

        ..  container:: example

            Accepts strings:

            >>> staff = abjad.Staff("c'4 d' e' f'")
            >>> score_block = abjad.Block(name="score")
            >>> score_block.items.append(staff)
            >>> lilypond_file = abjad.LilyPondFile(
            ...     lilypond_language_token=False,
            ...     lilypond_version_token=False,
            ... )
            >>> string = r"\customCommand"
            >>> lilypond_file.items.append(string)
            >>> lilypond_file.items.append(score_block)

            >>> string = abjad.lilypond(lilypond_file)
            >>> print(string)
            \customCommand
            <BLANKLINE>
            \score {
                \new Staff
                {
                    c'4
                    d'4
                    e'4
                    f'4
                }
            }

        """
        return self._items

    @property
    def layout_block(self) -> typing.Optional[Block]:
        """
        Gets layout block.

        ..  container:: example

            >>> block = abjad.Block(name="layout")
            >>> lilypond_file = abjad.LilyPondFile(items=[block])
            >>> lilypond_file.layout_block
            <Block(name='layout')>

        """
        for item in self.items:
            if isinstance(item, Block):
                if item.name == "layout":
                    return item
        return None

    @property
    def lilypond_language_token(self) -> typing.Optional[LilyPondLanguageToken]:
        """
        Gets LilyPond language token.

        ..  container:: example

            >>> lilypond_file = abjad.LilyPondFile()
            >>> lilypond_file.lilypond_language_token
            LilyPondLanguageToken()

        """
        return self._lilypond_language_token

    @property
    def lilypond_version_token(self) -> typing.Optional[LilyPondVersionToken]:
        """
        Gets LilyPond version token.

        ..  container:: example

            >>> lilypond_file = abjad.LilyPondFile()
            >>> lilypond_file.lilypond_version_token # doctest: +SKIP
            LilyPondVersionToken('2.19.35')

        """
        return self._lilypond_version_token

    @property
    def paper_block(self) -> typing.Optional[Block]:
        """
        Gets paper block.

        ..  container:: example

            Gets paper block:

            >>> block = abjad.Block(name="paper")
            >>> lilypond_file = abjad.LilyPondFile(items=[block])
            >>> lilypond_file.paper_block
            <Block(name='paper')>

        """
        for item in self.items:
            if isinstance(item, Block):
                if item.name == "paper":
                    return item
        return None

    @property
    def score_block(self) -> typing.Optional[Block]:
        """
        Gets score block.

        ..  container:: example

            >>> block = abjad.Block(name="score")
            >>> lilypond_file = abjad.LilyPondFile(items=[block])
            >>> lilypond_file.score_block
            <Block(name='score')>

        """
        for item in self.items:
            if isinstance(item, Block):
                if item.name == "score":
                    return item
        return None

    @property
    def use_relative_includes(self) -> typing.Optional[bool]:
        """
        Is true when LilyPond file should use relative includes.

        ..  container:: example

            >>> lilypond_file = abjad.LilyPondFile()
            >>> lilypond_file.use_relative_includes is None
            True

        """
        return self._use_relative_includes

    ### PUBLIC METHODS ###

    def get_tag(self, site=None):
        """
        Gets tag.
        """
        tag = _tag.Tag(self._tag)
        tag = tag.append(site)
        return tag

    @classmethod
    def rhythm(
        class_,
        selections,
        divisions=None,
        attach_lilypond_voice_commands=None,
        implicit_scaling=None,
        pitched_staff=None,
        simultaneous_selections=None,
        time_signatures=None,
    ) -> "LilyPondFile":
        r"""
        Makes rhythm-styled LilyPond file.

        ..  container:: example

            Makes rhythmic staff:

            >>> divisions = [(3, 4), (4, 8), (1, 4)]
            >>> maker = abjad.NoteMaker()
            >>> selections = [
            ...     maker(6 * [0], [(1, 8)]),
            ...     maker(8 * [0], [(1, 16)]),
            ...     maker(2 * [0], [(1, 8)]),
            ... ]
            >>> for selection in selections:
            ...     abjad.beam(selection[:])
            ...
            >>> lilypond_file = abjad.LilyPondFile.rhythm(
            ...     selections,
            ...     divisions,
            ...     )
            >>> abjad.illustrators.attach_markup_struts(lilypond_file)
            >>> abjad.show(lilypond_file) # doctest: +SKIP

            ..  docs::

                >>> score = lilypond_file[abjad.Score]
                >>> string = abjad.lilypond(score)
                >>> print(string)
                \new Score
                <<
                    \new GlobalContext
                    {
                        \time 3/4
                        s1 * 3/4
                        \time 4/8
                        s1 * 1/2
                        \time 1/4
                        s1 * 1/4
                    }
                    \new RhythmicStaff
                    {
                        c'8
                        - \tweak staff-padding 11
                        - \tweak transparent ##t
                        ^ \markup I
                        [
                        c'8
                        c'8
                        c'8
                        c'8
                        c'8
                        ]
                        c'16
                        [
                        c'16
                        c'16
                        c'16
                        c'16
                        c'16
                        c'16
                        c'16
                        ]
                        c'8
                        [
                        c'8
                        ]
                    }
                >>

        ..  container:: example

            Set time signatures explicitly:

            >>> divisions = [(3, 4), (4, 8), (1, 4)]
            >>> maker = abjad.NoteMaker()
            >>> selections = [
            ...     maker(6 * [0], [(1, 8)]),
            ...     maker(8 * [0], [(1, 16)]),
            ...     maker(2 * [0], [(1, 8)]),
            ... ]
            >>> for selection in selections:
            ...     abjad.beam(selection[:])
            ...
            >>> lilypond_file = abjad.LilyPondFile.rhythm(
            ...     selections,
            ...     [(6, 8), (4, 8), (2, 8)],
            ... )
            >>> abjad.illustrators.attach_markup_struts(lilypond_file)
            >>> abjad.show(lilypond_file) # doctest: +SKIP

            ..  docs::

                >>> score = lilypond_file[abjad.Score]
                >>> string = abjad.lilypond(score)
                >>> print(string)
                \new Score
                <<
                    \new GlobalContext
                    {
                        \time 6/8
                        s1 * 3/4
                        \time 4/8
                        s1 * 1/2
                        \time 2/8
                        s1 * 1/4
                    }
                    \new RhythmicStaff
                    {
                        c'8
                        - \tweak staff-padding 11
                        - \tweak transparent ##t
                        ^ \markup I
                        [
                        c'8
                        c'8
                        c'8
                        c'8
                        c'8
                        ]
                        c'16
                        [
                        c'16
                        c'16
                        c'16
                        c'16
                        c'16
                        c'16
                        c'16
                        ]
                        c'8
                        [
                        c'8
                        ]
                    }
                >>

        ..  container:: example

            Makes pitched staff:

            >>> divisions = [(3, 4), (4, 8), (1, 4)]
            >>> maker = abjad.NoteMaker()
            >>> selections = [
            ...     maker(6 * [0], [(1, 8)]),
            ...     maker(8 * [0], [(1, 16)]),
            ...     maker(2 * [0], [(1, 8)]),
            ... ]
            >>> for selection in selections:
            ...     abjad.beam(selection[:])
            ...
            >>> lilypond_file = abjad.LilyPondFile.rhythm(
            ...     selections,
            ...     divisions,
            ...     pitched_staff=True,
            ... )
            >>> abjad.illustrators.attach_markup_struts(lilypond_file)
            >>> abjad.show(lilypond_file) # doctest: +SKIP

            ..  docs::

                >>> string = abjad.lilypond(lilypond_file[abjad.Score])
                >>> print(string)
                \new Score
                <<
                    \new GlobalContext
                    {
                        \time 3/4
                        s1 * 3/4
                        \time 4/8
                        s1 * 1/2
                        \time 1/4
                        s1 * 1/4
                    }
                    \new Staff
                    {
                        c'8
                        - \tweak staff-padding 11
                        - \tweak transparent ##t
                        ^ \markup I
                        [
                        c'8
                        c'8
                        c'8
                        c'8
                        c'8
                        ]
                        c'16
                        [
                        c'16
                        c'16
                        c'16
                        c'16
                        c'16
                        c'16
                        c'16
                        ]
                        c'8
                        [
                        c'8
                        ]
                    }
                >>

        ..  container:: example

            Makes simultaneous voices:

            >>> divisions = [(3, 4), (4, 8), (1, 4)]
            >>> maker = abjad.NoteMaker()
            >>> selections = [
            ...     maker(6 * [0], [(1, 8)]),
            ...     maker(8 * [0], [(1, 16)]),
            ...     maker(2 * [0], [(1, 8)]),
            ... ]
            >>> for selection in selections:
            ...     abjad.beam(selection[:])
            ...
            >>> for note in abjad.iterate(selections).components(abjad.Note):
            ...     note.written_pitch = abjad.NamedPitch("e'")
            ...
            >>> selection_1 = selections[0] + selections[1] + selections[2]
            >>> selections = [
            ...     maker(12 * [0], [(1, 16)]),
            ...     maker(16 * [0], [(1, 32)]),
            ...     maker(4 * [0], [(1, 16)]),
            ... ]
            >>> for selection in selections:
            ...     abjad.beam(selection[:])
            ...
            >>> selection_2 = selections[0] + selections[1] + selections[2]
            >>> selections = {
            ...     "Voice_1": selection_1,
            ...     "Voice_2": selection_2,
            ... }
            >>> lilypond_file = abjad.LilyPondFile.rhythm(
            ...     selections,
            ...     divisions,
            ... )
            >>> abjad.illustrators.attach_markup_struts(lilypond_file)
            >>> voice_1 = lilypond_file["Voice_1"]
            >>> literal = abjad.LilyPondLiteral(r"\voiceOne", "opening")
            >>> abjad.attach(literal, voice_1)
            >>> voice_2 = lilypond_file["Voice_2"]
            >>> literal = abjad.LilyPondLiteral(r"\voiceTwo", "opening")
            >>> abjad.attach(literal, voice_2)
            >>> abjad.show(lilypond_file) # doctest: +SKIP

            ..  docs::

                >>> string = abjad.lilypond(lilypond_file[abjad.Score])
                >>> print(string)
                \new Score
                <<
                    \new GlobalContext
                    {
                        s1 * 3/4
                        s1 * 1/2
                        s1 * 1/4
                    }
                    \new Staff
                    <<
                        \context Voice = "Voice_1"
                        {
                            \voiceOne
                            e'8
                            - \tweak staff-padding 11
                            - \tweak transparent ##t
                            ^ \markup I
                            [
                            e'8
                            e'8
                            e'8
                            e'8
                            e'8
                            ]
                            e'16
                            [
                            e'16
                            e'16
                            e'16
                            e'16
                            e'16
                            e'16
                            e'16
                            ]
                            e'8
                            [
                            e'8
                            ]
                        }
                        \context Voice = "Voice_2"
                        {
                            \voiceTwo
                            c'16
                            [
                            c'16
                            c'16
                            c'16
                            c'16
                            c'16
                            c'16
                            c'16
                            c'16
                            c'16
                            c'16
                            c'16
                            ]
                            c'32
                            [
                            c'32
                            c'32
                            c'32
                            c'32
                            c'32
                            c'32
                            c'32
                            c'32
                            c'32
                            c'32
                            c'32
                            c'32
                            c'32
                            c'32
                            c'32
                            ]
                            c'16
                            [
                            c'16
                            c'16
                            c'16
                            ]
                        }
                    >>
                >>

        """
        if isinstance(selections, Selection):
            pass
        elif isinstance(selections, list):
            for selection in selections:
                if not isinstance(selection, Selection):
                    raise TypeError(f"must be selection: {selection!r}.")
        elif isinstance(selections, dict):
            for selection in selections.values():
                if not isinstance(selection, Selection):
                    raise TypeError(f"must be selection: {selection!r}.")
        else:
            raise TypeError(f"must be list or dictionary: {selections!r}.")
        score = Score()
        lilypond_file = class_(
            includes=["default.ily", "rhythm-maker-docs.ily"],
            items=[
                Block(name="header"),
                Block(name="layout"),
                Block(name="paper"),
                Block(name="score"),
            ],
        )
        assert lilypond_file.header_block is not None
        lilypond_file.header_block.tagline = "##f"
        assert lilypond_file.score_block is not None
        lilypond_file.score_block.items.append(score)
        if pitched_staff is None:
            if isinstance(selections, (list, Selection)):
                selections_ = selections
            elif isinstance(selections, dict):
                selections_ = list(selections.values())
            else:
                raise TypeError(selections)
            notes = Selection(selections_).notes()
            assert isinstance(notes, Selection)
            for note in notes:
                if note.written_pitch != NamedPitch("c'"):
                    pitched_staff = True
                    break
            chords = Selection(selections_).chords()
            if chords:
                pitched_staff = True
        if isinstance(selections, (list, Selection)):
            if divisions is None:
                duration = _inspect._get_duration(selections)
                divisions = [duration]
            time_signatures = time_signatures or divisions
            time_signatures = [TimeSignature(_) for _ in time_signatures]
            if pitched_staff:
                staff = Staff()
            else:
                staff = Staff(lilypond_type="RhythmicStaff")
            staff.extend(selections)
        elif isinstance(selections, dict):
            voices = []
            for voice_name in sorted(selections):
                selections_ = selections[voice_name]
                selections__ = Sequence(selections_).flatten(depth=-1)
                selections__ = copy.deepcopy(selections__)
                voice = Voice(selections__, name=voice_name)
                if attach_lilypond_voice_commands:
                    voice_name_to_command_string = {
                        "Voice_1": "voiceOne",
                        "Voice_2": "voiceTwo",
                        "Voice_3": "voiceThree",
                        "Voice_4": "voiceFour",
                    }
                    command_string = voice_name_to_command_string.get(voice_name)
                    if command_string:
                        command = LilyPondLiteral("\\" + command_string)
                        attach(command, voice)
                voices.append(voice)
            staff = Staff(voices, simultaneous=True)
            if divisions is None:
                duration = staff._get_duration()
                divisions = [duration]
        else:
            message = "must be list or dictionary of selections:"
            message += f" {selections!r}."
            raise TypeError(message)
        score.append(staff)
        assert isinstance(divisions, collections.abc.Sequence), repr(divisions)
        time_signatures = time_signatures or divisions
        context = Context(lilypond_type="GlobalContext")
        skips = []
        for time_signature in time_signatures:
            skip = Skip(1)
            skip.multiplier = time_signature
            attach(time_signature, skip, context="Score")
            skips.append(skip)
        context.extend(skips)
        score.insert(0, context)
        return lilypond_file



import abc
import collections
import types

from .. import enumerate, math
from ..storage import FormatSpecification
from ..typedcollections import TypedCollection, TypedCounter
from .intervalclasses import IntervalClass, NamedIntervalClass, NumberedIntervalClass
from .intervals import Interval, NamedInterval, NumberedInterval
from .pitchclasses import NamedPitchClass, NumberedPitchClass, PitchClass
from .pitches import NamedPitch, NumberedPitch, Pitch
from .segments import PitchClassSegment, PitchSegment
from .sets import PitchClassSet, PitchSet


class Vector(TypedCounter):
    """
    Abstract vector.
    """

    ### CLASS VARIABLES ##

    __slots__ = ()

    ### INITIALIZER ###

    def __init__(self, items=None, item_class=None):
        prototype_1 = (collections.abc.Iterator, types.GeneratorType)
        prototype_2 = (TypedCounter, collections.Counter)
        if isinstance(items, str):
            items = items.split()
        elif isinstance(items, prototype_1):
            items = [item for item in items]
        elif isinstance(items, dict):
            items = self._dictionary_to_items(items, item_class)
        if isinstance(items, prototype_2):
            new_tokens = []
            for item, count in items.items():
                new_tokens.extend(count * [item])
            items = new_tokens
        if item_class is None:
            item_class = self._named_item_class
            if items is not None:
                if isinstance(items, TypedCollection) and issubclass(
                    items.item_class, self._parent_item_class
                ):
                    item_class = items.item_class
                elif len(items):
                    if isinstance(items, collections.abc.Set):
                        items = tuple(items)
                    if isinstance(items, dict):
                        item_class = self._dictionary_to_item_class(items)
                    elif isinstance(items[0], str):
                        item_class = self._named_item_class
                    elif isinstance(items[0], (int, float)):
                        item_class = self._numbered_item_class
                    elif isinstance(items[0], self._parent_item_class):
                        item_class = type(items[0])
        assert issubclass(item_class, self._parent_item_class)
        TypedCounter.__init__(self, items=items, item_class=item_class)

    ### SPECIAL METHODS ###

    def __str__(self):
        """
        String representation of vector.

        Returns string.
        """
        parts = [f"{key}: {value}" for key, value in self.items()]
        string = ", ".join(parts)
        return f"<{string}>"

    ### PRIVATE PROPERTIES ###

    @abc.abstractproperty
    def _named_item_class(self):
        raise NotImplementedError

    @abc.abstractproperty
    def _numbered_item_class(self):
        raise NotImplementedError

    @abc.abstractproperty
    def _parent_item_class(self):
        raise NotImplementedError

    ### PRIVATE METHODS ###

    def _dictionary_to_item_class(self, dictionary):
        if not len(dictionary):
            return self._named_item_class
        keys = dictionary.keys()
        first_key = keys[0]
        assert isinstance(first_key, str), repr(first_key)
        try:
            float(first_key)
            item_class = self._numbered_item_class
        except ValueError:
            item_class = self._named_item_class
        return item_class

    def _dictionary_to_items(self, dictionary, item_class):
        items = []
        for initializer_token, count in dictionary.items():
            for _ in range(count):
                item = item_class(initializer_token)
                items.append(item)
        return items

    def _get_format_specification(self):
        if self.item_class.__name__.startswith("Named"):
            repr_items = {str(k): v for k, v in self.items()}
        else:
            repr_items = {
                math.integer_equivalent_number_to_integer(float(k.number)): v
                for k, v in self.items()
            }
        return FormatSpecification(
            client=self,
            repr_is_indented=False,
            repr_args_values=[repr_items],
            storage_format_args_values=[self._collection],
        )

    ### PUBLIC METHODS ###

    @abc.abstractmethod
    def from_selection(class_, selection, item_class=None):
        """
        Makes vector from ``selection``.

        Returns vector.
        """
        raise NotImplementedError


class IntervalVector(Vector):
    """
    Interval vector.

    ..  container:: example

        Initializes from pitch segment:

        >>> pitch_segment = abjad.PitchSegment(
        ...     items=[0, 11, 7, 4, 2, 9, 3, 8, 10, 1, 5, 6],
        ...     )
        >>> numbered_interval_vector = abjad.IntervalVector(
        ...     items=pitch_segment,
        ...     item_class=abjad.NumberedInterval,
        ...     )
        >>> for interval, count in sorted(numbered_interval_vector.items(),
        ...     key=lambda x: (x[0].direction_number, x[0].number)):
        ...     print(interval, count)
        ...
        -11 1
        -10 1
        -9 1
        -8 2
        -7 3
        -6 3
        -5 4
        -4 4
        -3 4
        -2 5
        -1 6
        +1 5
        +2 5
        +3 5
        +4 4
        +5 3
        +6 3
        +7 2
        +8 2
        +9 2
        +10 1

    """

    ### CLASS VARIABLES ###

    __slots__ = ()

    ### INITIALIZER ###

    def __init__(self, items=None, item_class=None):
        if isinstance(
            items,
            (
                PitchSegment,
                PitchSet,
                PitchClassSegment,
                PitchClassSet,
            ),
        ):
            intervals = []
            pairs = enumerate.yield_pairs(items)
            for first, second in pairs:
                intervals.append(second - first)
            items = intervals
        Vector.__init__(self, items=items, item_class=item_class)

    ### SPECIAL METHODS ###

    def __repr__(self):
        """
        Gets interpret representation of interval vector.

        ..  container:: example

            Gets interpreter representation of interval vector:

            >>> pitch_segment = abjad.PitchSegment(
            ...     items=[0, 11, 7, 4, 2, 9, 3, 8, 10, 1, 5, 6],
            ...     )
            >>> vector = abjad.IntervalVector(
            ...     items=pitch_segment,
            ...     item_class=abjad.NumberedInterval,
            ...     )

            >>> vector
            IntervalVector({-11: 1, -10: 1, -9: 1, -8: 2, -7: 3, -6: 3, -5: 4, -4: 4, -3: 4, -2: 5, -1: 6, 1: 5, 2: 5, 3: 5, 4: 4, 5: 3, 6: 3, 7: 2, 8: 2, 9: 2, 10: 1}, item_class=NumberedInterval)

        ..  container:: example

            Initializes from interpreter representation of interval vector:

            >>> abjad.IntervalVector(vector)
            IntervalVector({-11: 1, -10: 1, -9: 1, -8: 2, -7: 3, -6: 3, -5: 4, -4: 4, -3: 4, -2: 5, -1: 6, 1: 5, 2: 5, 3: 5, 4: 4, 5: 3, 6: 3, 7: 2, 8: 2, 9: 2, 10: 1}, item_class=NumberedInterval)

        Returns string.
        """
        return super().__repr__()

    ### PRIVATE PROPERTIES ###

    @property
    def _named_item_class(self):
        return NamedInterval

    @property
    def _numbered_item_class(self):
        return NumberedInterval

    @property
    def _parent_item_class(self):
        return Interval

    ### PUBLIC METHODS ###

    @classmethod
    def from_selection(class_, selection, item_class=None):
        """
        Makes interval vector from ``selection``.

        Returns interval vector.
        """
        pitch_segment = PitchSegment.from_selection(selection)
        return class_(pitch_segment, item_class=item_class)


class IntervalClassVector(Vector):
    """
    Interval-class vector.

    ..  container:: example

        An interval-class vector:

        >>> pitch_segment = abjad.PitchSegment(
        ...     items=[0, 11, 7, 4, 2, 9, 3, 8, 10, 1, 5, 6],
        ...     )
        >>> numbered_interval_class_vector = abjad.IntervalClassVector(
        ...     items=pitch_segment,
        ...     item_class=abjad.NumberedInversionEquivalentIntervalClass,
        ...     )

        >>> items = sorted(numbered_interval_class_vector.items())
        >>> for interval, count in items:
        ...     print(interval, count)
        ...
        1 12
        2 12
        3 12
        4 12
        5 12
        6 6

    """

    ### CLASS VARIABLES ###

    __slots__ = ()

    ### INITIALIZER ###

    def __init__(self, items=None, item_class=None):
        prototype = (
            PitchSegment,
            PitchSet,
            PitchClassSegment,
            PitchClassSet,
        )
        if isinstance(items, prototype):
            intervals = []
            items = tuple(items)
            pairs = enumerate.yield_pairs(items)
            for first, second in pairs:
                intervals.append(second - first)
            items = intervals
        Vector.__init__(self, items=items, item_class=item_class)

    ### SPECIAL METHODS ###

    def __repr__(self):
        """
        Gets interpreter representation of interval-class vector.

        ..  container:: example

            Gets interpreter representation of interval-class vector:

            >>> pitch_segment = abjad.PitchSegment(
            ...     items=[0, 11, 7, 4, 2, 9, 3, 8, 10, 1, 5, 6],
            ...     )
            >>> vector = abjad.IntervalClassVector(
            ...     items=pitch_segment,
            ...     item_class=abjad.NumberedInversionEquivalentIntervalClass,
            ...     )

            >>> vector
            IntervalClassVector({1: 12, 2: 12, 3: 12, 4: 12, 5: 12, 6: 6}, item_class=NumberedInversionEquivalentIntervalClass)

        ..  container:: example

            Initializes from interpreter representation of interval-class
            vector:

            >>> abjad.IntervalClassVector(vector)
            IntervalClassVector({1: 12, 2: 12, 3: 12, 4: 12, 5: 12, 6: 6}, item_class=NumberedInversionEquivalentIntervalClass)

        Returns string.
        """
        return super().__repr__()

    ### PRIVATE PROPERTIES ###

    @property
    def _label(self):
        counts = []
        for i in range(7):
            counts.append(self[i])
        counts = "".join([str(x) for x in counts])
        if len(self) == 13:
            quartertones = []
            for i in range(6):
                quartertones.append(self[i + 0.5])
            quartertones = "".join([str(x) for x in quartertones])
            return r'\tiny \column { "%s" "%s" }' % (counts, quartertones)
        else:
            return r"\tiny %s" % counts

    @property
    def _named_item_class(self):
        return NamedIntervalClass

    @property
    def _numbered_item_class(self):
        return NumberedIntervalClass

    @property
    def _parent_item_class(self):
        return IntervalClass

    ### PUBLIC METHODS ###

    @classmethod
    def from_selection(class_, selection, item_class=None):
        """
        Makes interval-class vector from ``selection``.

        ..  container:: example

            Makes numbered inversion-equivalent interval-class vector from
            selection:

            >>> chord = abjad.Chord("<c' d' b''>4"),
            >>> vector = abjad.IntervalClassVector.from_selection(
            ...     abjad.select(chord),
            ...     item_class=abjad.NumberedInversionEquivalentIntervalClass,
            ...     )
            >>> vector
            IntervalClassVector({1: 1, 2: 1, 3: 1}, item_class=NumberedInversionEquivalentIntervalClass)

        ..  container:: example

            Makes numbered interval-class vector from selection:

            >>> chord = abjad.Chord("<c' d' b''>4")
            >>> vector = abjad.IntervalClassVector.from_selection(
            ...     abjad.select(chord),
            ...     item_class=abjad.NumberedIntervalClass,
            ...     )
            >>> vector
            IntervalClassVector({-11: 1, -9: 1, -2: 1}, item_class=NumberedIntervalClass)

            .. todo:: This should probabaly be checked. Resulting values
                should probabaly be positive (or signless) instead of negative.

        ..  container:: example

            Makes named interval-class vector from selection:

            >>> chord = abjad.Chord("<c' d' b''>4")
            >>> vector = abjad.IntervalClassVector.from_selection(
            ...     abjad.select(chord),
            ...     item_class=None,
            ...     )
            >>> vector
            IntervalClassVector({'-M2': 1, '-M6': 1, '-M7': 1}, item_class=NamedIntervalClass)

            .. todo:: This should probabaly be checked. Resulting values
                should probabaly be positive (or signless) instead of negative.

        Returns new interval-class vector.
        """
        pitch_segment = PitchSegment.from_selection(selection)
        return class_(pitch_segment, item_class=item_class)


class PitchClassVector(Vector):
    """
    Pitch-class vector.

    ..  container:: example

        Pitch-class vector:

        >>> vector = abjad.PitchClassVector(
        ...     items=[7, 6, -2, -3, -3, 0, 1, 14, 15, 16, 16],
        ...     item_class=abjad.NumberedPitchClass,
        ...     )

        >>> items = sorted(vector.items())
        >>> for pitch_class, count in items:
        ...     print(pitch_class, count)
        0 1
        1 1
        2 1
        3 1
        4 2
        6 1
        7 1
        9 2
        10 1

    """

    ### CLASS VARIABLES ###

    __slots__ = ()

    ### SPECIAL METHODS ###

    def __repr__(self):
        """
        Gets interpreter representation of pitch-class vector.

        ..  container:: example

            Gets interpreter representation of pitch-class vector:

            >>> vector = abjad.PitchClassVector(
            ...     items=[7, 6, -2, -3, -3, 0, 1, 14, 15, 16, 16],
            ...     item_class=abjad.NumberedPitchClass,
            ...     )

            >>> vector
            PitchClassVector({0: 1, 1: 1, 2: 1, 3: 1, 4: 2, 6: 1, 7: 1, 9: 2, 10: 1}, item_class=NumberedPitchClass)

        ..  container:: example

            Initializes from interpreter representation of pitch-class vector:


                >>> abjad.PitchClassVector(vector)
                PitchClassVector({0: 1, 1: 1, 2: 1, 3: 1, 4: 2, 6: 1, 7: 1, 9: 2, 10: 1}, item_class=NumberedPitchClass)

        Returns string.
        """
        return super().__repr__()

    ### PRIVATE PROPERTIES ###

    @property
    def _named_item_class(self):
        return NamedPitchClass

    @property
    def _numbered_item_class(self):
        return NumberedPitchClass

    @property
    def _parent_item_class(self):
        return PitchClass

    ### PUBLIC METHODS ###

    @classmethod
    def from_selection(class_, selection, item_class=None):
        """
        Makes pitch-class vector from ``selection``.

        Returns pitch-class vector.
        """
        pitch_segment = PitchSegment.from_selection(selection)
        return class_(pitch_segment, item_class=item_class)


class PitchVector(Vector):
    """
    Pitch vector.

    ..  container:: example

        >>> vector = abjad.PitchVector(
        ...     items=[7, 6, -2, -3, -3, 0, 1, 14, 15, 16, 16],
        ...     item_class=abjad.NumberedPitch,
        ...     )

        >>> items = list(vector.items())
        >>> items.sort(key=lambda x: x[0].number)
        >>> for pitch_class, count in items:
        ...     print(pitch_class, count)
        -3 2
        -2 1
        0 1
        1 1
        6 1
        7 1
        14 1
        15 1
        16 2

    """

    ### CLASS VARIABLES ###

    __slots__ = ()

    ### SPECIAL METHODS ###

    def __repr__(self):
        """
        Gets interpreter representation of pitch vector.

        ..  container:: example

            Gets interpreter representation of pitch vector:

            >>> vector = abjad.PitchVector(
            ...     items=[7, 6, -2, -3, -3, 0, 1, 14, 15, 16, 16],
            ...     item_class=abjad.NumberedPitch,
            ...     )

            >>> vector
            PitchVector({-3: 2, -2: 1, 0: 1, 1: 1, 6: 1, 7: 1, 14: 1, 15: 1, 16: 2}, item_class=NumberedPitch)

        ..  container:: example

            Initializes from interpreter representation of pitch vector:

                >>> abjad.PitchVector(vector)
                PitchVector({-3: 2, -2: 1, 0: 1, 1: 1, 6: 1, 7: 1, 14: 1, 15: 1, 16: 2}, item_class=NumberedPitch)

        Returns string.
        """
        return super().__repr__()

    ### PRIVATE PROPERTIES ###

    @property
    def _named_item_class(self):
        return NamedPitch

    @property
    def _numbered_item_class(self):
        return NumberedPitch

    @property
    def _parent_item_class(self):
        return Pitch

    ### PUBLIC METHODS ###

    @classmethod
    def from_selection(class_, selection, item_class=None):
        """
        Makes pitch vector from ``selection``.

        Returns pitch vector.
        """
        pitch_segment = PitchSegment.from_selection(selection)
        return class_(pitch_segment, item_class=item_class)



import abc
import collections
import typing

from . import _inspect, _iterate, instruments
from .attach import Wrapper, annotate, attach
from .illustrators import illustrate
from .indicators.Clef import Clef
from .instruments import Piano
from .iterate import Iteration
from .lilypondfile import LilyPondFile
from .new import new
from .ordereddict import OrderedDict
from .overrides import LilyPondLiteral
from .score import Context, Score, Skip, Staff, StaffGroup, Voice
from .select import Selection
from .storage import StorageFormatManager
from .tag import Tag


class ScoreTemplate:
    """
    Abstract score template.
    """

    ### CLASS VARIABLES ###

    __documentation_section__: typing.Optional[str] = "Score templates"

    __slots__ = ("_voice_abbreviations",)

    _always_make_global_rests = False

    _do_not_require_margin_markup = False

    ### INITIALIZER ###

    def __init__(self):
        self._voice_abbreviations = OrderedDict()

    ### SPECIAL METHODS ###

    @abc.abstractmethod
    def __call__(self) -> Score:
        """
        Calls score template.
        """
        pass

    def __illustrate__(
        self, default_paper_size=None, global_staff_size=None, includes=None
    ):
        """
        Illustrates score template.
        """
        score: Score = self()
        site = "abjad.ScoreTemplate.__illustrate__()"
        tag = Tag(site)
        for voice in Iteration(score).components(Voice):
            skip = Skip(1, tag=tag)
            voice.append(skip)
        self.attach_defaults(score)
        lilypond_file: LilyPondFile = illustrate(score)
        lilypond_file = new(
            lilypond_file,
            default_paper_size=default_paper_size,
            global_staff_size=global_staff_size,
            includes=includes,
        )
        return lilypond_file

    def __repr__(self) -> str:
        """
        Gets interpreter representation.
        """
        return StorageFormatManager(self).get_repr_format()

    ### PRIVATE METHODS ###

    def _make_global_context(self):
        site = "abjad.ScoreTemplate._make_global_context()"
        tag = Tag(site)
        global_rests = Context(
            lilypond_type="GlobalRests",
            name="Global_Rests",
            tag=tag,
        )
        global_skips = Context(
            lilypond_type="GlobalSkips",
            name="Global_Skips",
            tag=tag,
        )
        global_context = Context(
            [global_rests, global_skips],
            lilypond_type="GlobalContext",
            simultaneous=True,
            name="Global_Context",
            tag=tag,
        )
        return global_context

    ### PUBLIC PROPERTIES ###

    @property
    def always_make_global_rests(self) -> bool:
        """
        Is true when score template always makes global rests.
        """
        return self._always_make_global_rests

    @property
    def do_not_require_margin_markup(self) -> bool:
        """
        Is true when score template does not require margin markup.

        Conventionally, solos do not require margin markup.
        """
        return self._do_not_require_margin_markup

    ### PUBLIC METHODS ###

    def allows_instrument(
        self, staff_name: str, instrument: instruments.Instrument
    ) -> bool:
        """
        Is true when ``staff_name`` allows ``instrument``.

        To be implemented by concrete score template classes.
        """
        return True

    def attach_defaults(self, argument) -> typing.List:
        """
        Attaches defaults to all staff and staff group contexts in
        ``argument`` when ``argument`` is a score.

        Attaches defaults to ``argument`` (without iterating ``argument``) when
        ``argument`` is a staff or staff group.

        Returns list of one wrapper for every indicator attached.
        """
        assert isinstance(argument, (Score, Staff, StaffGroup)), repr(argument)
        wrappers: typing.List[Wrapper] = []
        prototype = (Staff, StaffGroup)
        staves = Selection(argument).components(prototype)
        assert isinstance(staves, Selection), repr(staves)
        for staff in staves:
            leaf = _iterate._get_leaf(staff, 0)
            clef = _inspect._get_indicator(leaf, Clef)
            if clef is not None:
                continue
            clef = _inspect._get_annotation(staff, "default_clef")
            if clef is not None:
                wrapper = attach(
                    clef,
                    leaf,
                    tag=Tag("abjad.ScoreTemplate.attach_defaults(3)"),
                    wrapper=True,
                )
                wrappers.append(wrapper)
        return wrappers

    ### PUBLIC PROPERTIES ###

    @property
    def voice_abbreviations(self) -> OrderedDict:
        """
        Gets voice abbreviations.
        """
        return self._voice_abbreviations


class GroupedRhythmicStavesScoreTemplate(ScoreTemplate):
    r"""
    Grouped rhythmic staves score template.

    ..  container:: example

        One voice per staff:

        >>> class_ = abjad.GroupedRhythmicStavesScoreTemplate
        >>> template_1 = class_(staff_count=4)
        >>> abjad.show(template_1) # doctest: +SKIP

        ..  docs::

            >>> score = template_1.__illustrate__()[abjad.Score]
            >>> string = abjad.lilypond(score)
            >>> print(string)
            \context Score = "Grouped_Rhythmic_Staves_Score"
            <<
                \context StaffGroup = "Grouped_Rhythmic_Staves_Staff_Group"
                <<
                    \context RhythmicStaff = "Staff_1"
                    {
                        \context Voice = "Voice_1"
                        {
                            \clef "percussion"
                            s1
                        }
                    }
                    \context RhythmicStaff = "Staff_2"
                    {
                        \context Voice = "Voice_2"
                        {
                            \clef "percussion"
                            s1
                        }
                    }
                    \context RhythmicStaff = "Staff_3"
                    {
                        \context Voice = "Voice_3"
                        {
                            \clef "percussion"
                            s1
                        }
                    }
                    \context RhythmicStaff = "Staff_4"
                    {
                        \context Voice = "Voice_4"
                        {
                            \clef "percussion"
                            s1
                        }
                    }
                >>
            >>

        >>> score = template_1()
        >>> abjad.show(score) # doctest: +SKIP

        >>> string = abjad.lilypond(score)
        >>> print(string)
        \context Score = "Grouped_Rhythmic_Staves_Score"
        <<
            \context StaffGroup = "Grouped_Rhythmic_Staves_Staff_Group"
            <<
                \context RhythmicStaff = "Staff_1"
                {
                    \context Voice = "Voice_1"
                    {
                    }
                }
                \context RhythmicStaff = "Staff_2"
                {
                    \context Voice = "Voice_2"
                    {
                    }
                }
                \context RhythmicStaff = "Staff_3"
                {
                    \context Voice = "Voice_3"
                    {
                    }
                }
                \context RhythmicStaff = "Staff_4"
                {
                    \context Voice = "Voice_4"
                    {
                    }
                }
            >>
        >>

    ..  container:: example

        More than one voice per staff:

        >>> template_2 = class_(staff_count=[2, 1, 2])
        >>> abjad.show(template_2) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(template_2.__illustrate__()[abjad.Score])
            >>> print(string)
            \context Score = "Grouped_Rhythmic_Staves_Score"
            <<
                \context StaffGroup = "Grouped_Rhythmic_Staves_Staff_Group"
                <<
                    \context RhythmicStaff = "Staff_1"
                    <<
                        \context Voice = "Voice_1_1"
                        {
                            s1
                        }
                        \context Voice = "Voice_1_2"
                        {
                            s1
                        }
                    >>
                    \context RhythmicStaff = "Staff_2"
                    {
                        \context Voice = "Voice_2"
                        {
                            s1
                        }
                    }
                    \context RhythmicStaff = "Staff_3"
                    <<
                        \context Voice = "Voice_3_1"
                        {
                            s1
                        }
                        \context Voice = "Voice_3_2"
                        {
                            s1
                        }
                    >>
                >>
            >>

        >>> score = template_2()
        >>> abjad.show(score) # doctest: +SKIP

        >>> string = abjad.lilypond(score)
        >>> print(string)
        \context Score = "Grouped_Rhythmic_Staves_Score"
        <<
            \context StaffGroup = "Grouped_Rhythmic_Staves_Staff_Group"
            <<
                \context RhythmicStaff = "Staff_1"
                <<
                    \context Voice = "Voice_1_1"
                    {
                    }
                    \context Voice = "Voice_1_2"
                    {
                    }
                >>
                \context RhythmicStaff = "Staff_2"
                {
                    \context Voice = "Voice_2"
                    {
                    }
                }
                \context RhythmicStaff = "Staff_3"
                <<
                    \context Voice = "Voice_3_1"
                    {
                    }
                    \context Voice = "Voice_3_2"
                    {
                    }
                >>
            >>
        >>

    """

    ### CLASS VARIABLES ###

    __slots__ = ("_staff_count",)

    ### INITIALIZER ###

    def __init__(self, staff_count=2):
        super().__init__()
        assert isinstance(staff_count, (int, collections.abc.Iterable))
        if isinstance(staff_count, collections.abc.Iterable):
            staff_count = list(staff_count)
        self._staff_count = staff_count

    ### SPECIAL METHODS ###

    def __call__(self):
        """
        Calls score template.

        Returns score.
        """
        staves = []
        site = "abjad.GroupedRhythmicStavesScoreTemplate.__call__()"
        tag = Tag(site)
        if isinstance(self.staff_count, int):
            for index in range(self.staff_count):
                number = index + 1
                name = f"Voice_{number}"
                voice = Voice([], name=name, tag=tag)
                name = f"Staff_{number}"
                staff = Staff([voice], name=name, tag=tag)
                staff.lilypond_type = "RhythmicStaff"
                annotate(staff, "default_clef", Clef("percussion"))
                staves.append(staff)
                key = f"v{number}"
                self.voice_abbreviations[key] = voice.name
        elif isinstance(self.staff_count, list):
            for staff_index, voice_count in enumerate(self.staff_count):
                staff_number = staff_index + 1
                name = f"Staff_{staff_number}"
                staff = Staff(name=name, tag=tag)
                staff.lilypond_type = "RhythmicStaff"
                assert 1 <= voice_count
                for voice_index in range(voice_count):
                    voice_number = voice_index + 1
                    if voice_count == 1:
                        voice_identifier = str(staff_number)
                    else:
                        voice_identifier = f"{staff_number}_{voice_number}"
                        staff.simultaneous = True
                    name = f"Voice_{voice_identifier}"
                    voice = Voice([], name=name, tag=tag)
                    staff.append(voice)
                    key = f"v{voice_identifier}"
                    self.voice_abbreviations[key] = voice.name
                staves.append(staff)
        grouped_rhythmic_staves_staff_group = StaffGroup(
            staves, name="Grouped_Rhythmic_Staves_Staff_Group", tag=tag
        )
        grouped_rhythmic_staves_score = Score(
            [grouped_rhythmic_staves_staff_group],
            name="Grouped_Rhythmic_Staves_Score",
            tag=tag,
        )
        return grouped_rhythmic_staves_score

    ### PUBLIC PROPERTIES ###

    @property
    def staff_count(self):
        """
        Gets score template staff count.

        ..  container:: example

            >>> class_ = abjad.GroupedRhythmicStavesScoreTemplate
            >>> template = class_(staff_count=4)
            >>> template.staff_count
            4

        Returns nonnegative integer.
        """
        return self._staff_count

    @property
    def voice_abbreviations(self):
        """
        Gets context name abbreviations.

        ..  container:: example

            >>> class_ = abjad.GroupedRhythmicStavesScoreTemplate
            >>> template = class_(staff_count=4)
            >>> template.voice_abbreviations
            OrderedDict([])

        """
        return super(GroupedRhythmicStavesScoreTemplate, self).voice_abbreviations


class GroupedStavesScoreTemplate(ScoreTemplate):
    r"""
    Grouped staves score template.

    ..  container:: example

        >>> class_ = abjad.GroupedStavesScoreTemplate
        >>> template = class_(staff_count=4)
        >>> abjad.show(template) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(template.__illustrate__()[abjad.Score])
            >>> print(string)
            \context Score = "Grouped_Staves_Score"
            <<
                \context StaffGroup = "Grouped_Staves_Staff_Group"
                <<
                    \context Staff = "Staff_1"
                    {
                        \context Voice = "Voice_1"
                        {
                            s1
                        }
                    }
                    \context Staff = "Staff_2"
                    {
                        \context Voice = "Voice_2"
                        {
                            s1
                        }
                    }
                    \context Staff = "Staff_3"
                    {
                        \context Voice = "Voice_3"
                        {
                            s1
                        }
                    }
                    \context Staff = "Staff_4"
                    {
                        \context Voice = "Voice_4"
                        {
                            s1
                        }
                    }
                >>
            >>

        >>> score = template()
        >>> string = abjad.lilypond(score)
        >>> print(string)
        \context Score = "Grouped_Staves_Score"
        <<
            \context StaffGroup = "Grouped_Staves_Staff_Group"
            <<
                \context Staff = "Staff_1"
                {
                    \context Voice = "Voice_1"
                    {
                    }
                }
                \context Staff = "Staff_2"
                {
                    \context Voice = "Voice_2"
                    {
                    }
                }
                \context Staff = "Staff_3"
                {
                    \context Voice = "Voice_3"
                    {
                    }
                }
                \context Staff = "Staff_4"
                {
                    \context Voice = "Voice_4"
                    {
                    }
                }
            >>
        >>

    """

    ### CLASS VARIABLES ###

    __slots__ = ("_staff_count",)

    ### INITIALIZER ###

    def __init__(self, staff_count=2):
        super().__init__()
        self._staff_count = staff_count

    ### SPECIAL METHODS ###

    def __call__(self):
        """
        Calls score template.

        Returns score.
        """
        staves = []
        site = "abjad.GroupedStavesScoreTemplate.__call__()"
        tag = Tag(site)
        for index in range(self.staff_count):
            number = index + 1
            voice = Voice([], name=f"Voice_{number}", tag=tag)
            staff = Staff([voice], name=f"Staff_{number}", tag=tag)
            staves.append(staff)
            self.voice_abbreviations[f"v{number}"] = voice.name
        staff_group = StaffGroup(staves, name="Grouped_Staves_Staff_Group", tag=tag)
        score = Score([staff_group], name="Grouped_Staves_Score", tag=tag)
        return score

    ### PUBLIC PROPERTIES ###

    @property
    def staff_count(self):
        """
        Gets staff count.

        ..  container::

            >>> class_ = abjad.GroupedStavesScoreTemplate
            >>> template = class_(staff_count=4)
            >>> template.staff_count
            4

        """
        return self._staff_count

    @property
    def voice_abbreviations(self):
        """
        Gets context name abbreviations.

        ..  container::

            >>> class_ = abjad.GroupedStavesScoreTemplate
            >>> template = class_(staff_count=4)
            >>> template.voice_abbreviations
            OrderedDict([])

        """
        return super().voice_abbreviations


class StringOrchestraScoreTemplate(ScoreTemplate):
    r"""
    String orchestra score template.

    ..  container:: example

        >>> template = abjad.StringOrchestraScoreTemplate()
        >>> abjad.show(template) # doctest: +SKIP

        >>> string = abjad.lilypond(template.__illustrate__()[abjad.Score])
        >>> print(string)
        \context Score = "Score"
        <<
            \tag #'(Violin_1 Violin_2 Violin_3 Violin_4 Violin_5 Violin_6 Viola_1 Viola_2 Viola_3 Viola_4 Cello_1 Cello_2 Cello_3 Contrabass_1 Contrabass_2)
            \context GlobalContext = "Global_Context"
            {
            }
            \context StaffGroup = "Outer_Staff_Group"
            <<
                \context ViolinStaffGroup = "Violin_Staff_Group"
                <<
                    \tag #'Violin_1
                    \context StringPerformerStaffGroup = "Violin_1_Staff_Group"
                    <<
                        \context BowingStaff = "Violin_1_Bowing_Staff"
                        <<
                            \context BowingVoice = "Violin_1_Bowing_Voice"
                            {
                                s1
                            }
                        >>
                        \context FingeringStaff = "Violin_1_Fingering_Staff"
                        <<
                            \context FingeringVoice = "Violin_1_Fingering_Voice"
                            {
                                \clef "treble"
                                s1
                            }
                        >>
                    >>
                    \tag #'Violin_2
                    \context StringPerformerStaffGroup = "Violin_2_Staff_Group"
                    <<
                        \context BowingStaff = "Violin_2_Bowing_Staff"
                        <<
                            \context BowingVoice = "Violin_2_Bowing_Voice"
                            {
                                s1
                            }
                        >>
                        \context FingeringStaff = "Violin_2_Fingering_Staff"
                        <<
                            \context FingeringVoice = "Violin_2_Fingering_Voice"
                            {
                                \clef "treble"
                                s1
                            }
                        >>
                    >>
                    \tag #'Violin_3
                    \context StringPerformerStaffGroup = "Violin_3_Staff_Group"
                    <<
                        \context BowingStaff = "Violin_3_Bowing_Staff"
                        <<
                            \context BowingVoice = "Violin_3_Bowing_Voice"
                            {
                                s1
                            }
                        >>
                        \context FingeringStaff = "Violin_3_Fingering_Staff"
                        <<
                            \context FingeringVoice = "Violin_3_Fingering_Voice"
                            {
                                \clef "treble"
                                s1
                            }
                        >>
                    >>
                    \tag #'Violin_4
                    \context StringPerformerStaffGroup = "Violin_4_Staff_Group"
                    <<
                        \context BowingStaff = "Violin_4_Bowing_Staff"
                        <<
                            \context BowingVoice = "Violin_4_Bowing_Voice"
                            {
                                s1
                            }
                        >>
                        \context FingeringStaff = "Violin_4_Fingering_Staff"
                        <<
                            \context FingeringVoice = "Violin_4_Fingering_Voice"
                            {
                                \clef "treble"
                                s1
                            }
                        >>
                    >>
                    \tag #'Violin_5
                    \context StringPerformerStaffGroup = "Violin_5_Staff_Group"
                    <<
                        \context BowingStaff = "Violin_5_Bowing_Staff"
                        <<
                            \context BowingVoice = "Violin_5_Bowing_Voice"
                            {
                                s1
                            }
                        >>
                        \context FingeringStaff = "Violin_5_Fingering_Staff"
                        <<
                            \context FingeringVoice = "Violin_5_Fingering_Voice"
                            {
                                \clef "treble"
                                s1
                            }
                        >>
                    >>
                    \tag #'Violin_6
                    \context StringPerformerStaffGroup = "Violin_6_Staff_Group"
                    <<
                        \context BowingStaff = "Violin_6_Bowing_Staff"
                        <<
                            \context BowingVoice = "Violin_6_Bowing_Voice"
                            {
                                s1
                            }
                        >>
                        \context FingeringStaff = "Violin_6_Fingering_Staff"
                        <<
                            \context FingeringVoice = "Violin_6_Fingering_Voice"
                            {
                                \clef "treble"
                                s1
                            }
                        >>
                    >>
                >>
                \context ViolaStaffGroup = "Viola_Staff_Group"
                <<
                    \tag #'Viola_1
                    \context StringPerformerStaffGroup = "Viola_1_Staff_Group"
                    <<
                        \context BowingStaff = "Viola_1_Bowing_Staff"
                        <<
                            \context BowingVoice = "Viola_1_Bowing_Voice"
                            {
                                s1
                            }
                        >>
                        \context FingeringStaff = "Viola_1_Fingering_Staff"
                        <<
                            \context FingeringVoice = "Viola_1_Fingering_Voice"
                            {
                                \clef "alto"
                                s1
                            }
                        >>
                    >>
                    \tag #'Viola_2
                    \context StringPerformerStaffGroup = "Viola_2_Staff_Group"
                    <<
                        \context BowingStaff = "Viola_2_Bowing_Staff"
                        <<
                            \context BowingVoice = "Viola_2_Bowing_Voice"
                            {
                                s1
                            }
                        >>
                        \context FingeringStaff = "Viola_2_Fingering_Staff"
                        <<
                            \context FingeringVoice = "Viola_2_Fingering_Voice"
                            {
                                \clef "alto"
                                s1
                            }
                        >>
                    >>
                    \tag #'Viola_3
                    \context StringPerformerStaffGroup = "Viola_3_Staff_Group"
                    <<
                        \context BowingStaff = "Viola_3_Bowing_Staff"
                        <<
                            \context BowingVoice = "Viola_3_Bowing_Voice"
                            {
                                s1
                            }
                        >>
                        \context FingeringStaff = "Viola_3_Fingering_Staff"
                        <<
                            \context FingeringVoice = "Viola_3_Fingering_Voice"
                            {
                                \clef "alto"
                                s1
                            }
                        >>
                    >>
                    \tag #'Viola_4
                    \context StringPerformerStaffGroup = "Viola_4_Staff_Group"
                    <<
                        \context BowingStaff = "Viola_4_Bowing_Staff"
                        <<
                            \context BowingVoice = "Viola_4_Bowing_Voice"
                            {
                                s1
                            }
                        >>
                        \context FingeringStaff = "Viola_4_Fingering_Staff"
                        <<
                            \context FingeringVoice = "Viola_4_Fingering_Voice"
                            {
                                \clef "alto"
                                s1
                            }
                        >>
                    >>
                >>
                \context CelloStaffGroup = "Cello_Staff_Group"
                <<
                    \tag #'Cello_1
                    \context StringPerformerStaffGroup = "Cello_1_Staff_Group"
                    <<
                        \context BowingStaff = "Cello_1_Bowing_Staff"
                        <<
                            \context BowingVoice = "Cello_1_Bowing_Voice"
                            {
                                s1
                            }
                        >>
                        \context FingeringStaff = "Cello_1_Fingering_Staff"
                        <<
                            \context FingeringVoice = "Cello_1_Fingering_Voice"
                            {
                                \clef "bass"
                                s1
                            }
                        >>
                    >>
                    \tag #'Cello_2
                    \context StringPerformerStaffGroup = "Cello_2_Staff_Group"
                    <<
                        \context BowingStaff = "Cello_2_Bowing_Staff"
                        <<
                            \context BowingVoice = "Cello_2_Bowing_Voice"
                            {
                                s1
                            }
                        >>
                        \context FingeringStaff = "Cello_2_Fingering_Staff"
                        <<
                            \context FingeringVoice = "Cello_2_Fingering_Voice"
                            {
                                \clef "bass"
                                s1
                            }
                        >>
                    >>
                    \tag #'Cello_3
                    \context StringPerformerStaffGroup = "Cello_3_Staff_Group"
                    <<
                        \context BowingStaff = "Cello_3_Bowing_Staff"
                        <<
                            \context BowingVoice = "Cello_3_Bowing_Voice"
                            {
                                s1
                            }
                        >>
                        \context FingeringStaff = "Cello_3_Fingering_Staff"
                        <<
                            \context FingeringVoice = "Cello_3_Fingering_Voice"
                            {
                                \clef "bass"
                                s1
                            }
                        >>
                    >>
                >>
                \context ContrabassStaffGroup = "Contrabass_Staff_Group"
                <<
                    \tag #'Contrabass_1
                    \context StringPerformerStaffGroup = "Contrabass_1_Staff_Group"
                    <<
                        \context BowingStaff = "Contrabass_1_Bowing_Staff"
                        <<
                            \context BowingVoice = "Contrabass_1_Bowing_Voice"
                            {
                                s1
                            }
                        >>
                        \context FingeringStaff = "Contrabass_1_Fingering_Staff"
                        <<
                            \context FingeringVoice = "Contrabass_1_Fingering_Voice"
                            {
                                \clef "bass_8"
                                s1
                            }
                        >>
                    >>
                    \tag #'Contrabass_2
                    \context StringPerformerStaffGroup = "Contrabass_2_Staff_Group"
                    <<
                        \context BowingStaff = "Contrabass_2_Bowing_Staff"
                        <<
                            \context BowingVoice = "Contrabass_2_Bowing_Voice"
                            {
                                s1
                            }
                        >>
                        \context FingeringStaff = "Contrabass_2_Fingering_Staff"
                        <<
                            \context FingeringVoice = "Contrabass_2_Fingering_Voice"
                            {
                                \clef "bass_8"
                                s1
                            }
                        >>
                    >>
                >>
            >>
        >>

    ..  container:: example

        As a string quartet:

        >>> template = abjad.StringOrchestraScoreTemplate(
        ...     violin_count=2,
        ...     viola_count=1,
        ...     cello_count=1,
        ...     contrabass_count=0,
        ...     )
        >>> abjad.show(template) # doctest: +SKIP

        >>> string = abjad.lilypond(template.__illustrate__()[abjad.Score])
        >>> print(string)
        \context Score = "Score"
        <<
            \tag #'(Violin_1 Violin_2 Viola Cello)
            \context GlobalContext = "Global_Context"
            {
            }
            \context StaffGroup = "Outer_Staff_Group"
            <<
                \context ViolinStaffGroup = "Violin_Staff_Group"
                <<
                    \tag #'Violin_1
                    \context StringPerformerStaffGroup = "Violin_1_Staff_Group"
                    <<
                        \context BowingStaff = "Violin_1_Bowing_Staff"
                        <<
                            \context BowingVoice = "Violin_1_Bowing_Voice"
                            {
                                s1
                            }
                        >>
                        \context FingeringStaff = "Violin_1_Fingering_Staff"
                        <<
                            \context FingeringVoice = "Violin_1_Fingering_Voice"
                            {
                                \clef "treble"
                                s1
                            }
                        >>
                    >>
                    \tag #'Violin_2
                    \context StringPerformerStaffGroup = "Violin_2_Staff_Group"
                    <<
                        \context BowingStaff = "Violin_2_Bowing_Staff"
                        <<
                            \context BowingVoice = "Violin_2_Bowing_Voice"
                            {
                                s1
                            }
                        >>
                        \context FingeringStaff = "Violin_2_Fingering_Staff"
                        <<
                            \context FingeringVoice = "Violin_2_Fingering_Voice"
                            {
                                \clef "treble"
                                s1
                            }
                        >>
                    >>
                >>
                \context ViolaStaffGroup = "Viola_Staff_Group"
                <<
                    \tag #'Viola
                    \context StringPerformerStaffGroup = "Viola_Staff_Group"
                    <<
                        \context BowingStaff = "Viola_Bowing_Staff"
                        <<
                            \context BowingVoice = "Viola_Bowing_Voice"
                            {
                                s1
                            }
                        >>
                        \context FingeringStaff = "Viola_Fingering_Staff"
                        <<
                            \context FingeringVoice = "Viola_Fingering_Voice"
                            {
                                \clef "alto"
                                s1
                            }
                        >>
                    >>
                >>
                \context CelloStaffGroup = "Cello_Staff_Group"
                <<
                    \tag #'Cello
                    \context StringPerformerStaffGroup = "Cello_Staff_Group"
                    <<
                        \context BowingStaff = "Cello_Bowing_Staff"
                        <<
                            \context BowingVoice = "Cello_Bowing_Voice"
                            {
                                s1
                            }
                        >>
                        \context FingeringStaff = "Cello_Fingering_Staff"
                        <<
                            \context FingeringVoice = "Cello_Fingering_Voice"
                            {
                                \clef "bass"
                                s1
                            }
                        >>
                    >>
                >>
            >>
        >>

    ..  container:: example

        As a cello solo:

        >>> template = abjad.StringOrchestraScoreTemplate(
        ...     violin_count=0,
        ...     viola_count=0,
        ...     cello_count=1,
        ...     contrabass_count=0,
        ...     )
        >>> abjad.show(template) # doctest: +SKIP

        >>> string = abjad.lilypond(template.__illustrate__()[abjad.Score])
        >>> print(string)
        \context Score = "Score"
        <<
            \tag #'(Cello)
            \context GlobalContext = "Global_Context"
            {
            }
            \context StaffGroup = "Outer_Staff_Group"
            <<
                \context CelloStaffGroup = "Cello_Staff_Group"
                <<
                    \tag #'Cello
                    \context StringPerformerStaffGroup = "Cello_Staff_Group"
                    <<
                        \context BowingStaff = "Cello_Bowing_Staff"
                        <<
                            \context BowingVoice = "Cello_Bowing_Voice"
                            {
                                s1
                            }
                        >>
                        \context FingeringStaff = "Cello_Fingering_Staff"
                        <<
                            \context FingeringVoice = "Cello_Fingering_Voice"
                            {
                                \clef "bass"
                                s1
                            }
                        >>
                    >>
                >>
            >>
        >>

    """

    ### CLASS VARIABLES ###

    __slots__ = (
        "_cello_count",
        "_contrabass_count",
        "_split_hands",
        "_use_percussion_clefs",
        "_viola_count",
        "_violin_count",
    )

    ### INITIALIZER ###

    def __init__(
        self,
        violin_count=6,
        viola_count=4,
        cello_count=3,
        contrabass_count=2,
        split_hands=True,
        use_percussion_clefs=False,
    ):
        assert 0 <= violin_count
        assert 0 <= viola_count
        assert 0 <= cello_count
        assert 0 <= contrabass_count
        super().__init__()
        self._violin_count = int(violin_count)
        self._viola_count = int(viola_count)
        self._cello_count = int(cello_count)
        self._contrabass_count = int(contrabass_count)
        self._split_hands = bool(split_hands)
        self._use_percussion_clefs = bool(use_percussion_clefs)

    ### SPECIAL METHODS ###

    def __call__(self):
        """
        Calls string orchestra template.

        Returns score.
        """
        site = "abjad.StringOrchestraScoreTemplate.__call__()"
        tag = Tag(site)

        ### TAGS ###

        tag_names = []

        ### SCORE ###

        staff_group = StaffGroup(name="Outer_Staff_Group", tag=tag)

        score = Score([staff_group], name="Score", tag=tag)

        ### VIOLINS ###

        if self.violin_count:
            clef_name = "treble"
            if self.use_percussion_clefs:
                clef_name = "percussion"
            instrument = instruments.Violin()
            instrument_count = self.violin_count
            (
                instrument_staff_group,
                instrument_tag_names,
            ) = self._make_instrument_staff_group(
                clef_name=clef_name,
                count=instrument_count,
                instrument=instrument,
            )
            staff_group.append(instrument_staff_group)
            tag_names.extend(instrument_tag_names)

        ### VIOLAS ###

        if self.viola_count:
            clef_name = "alto"
            if self.use_percussion_clefs:
                clef_name = "percussion"
            instrument = instruments.Viola()
            instrument_count = self.viola_count
            (
                instrument_staff_group,
                instrument_tag_names,
            ) = self._make_instrument_staff_group(
                clef_name=clef_name,
                count=instrument_count,
                instrument=instrument,
            )
            staff_group.append(instrument_staff_group)
            tag_names.extend(instrument_tag_names)

        ### CELLOS ###

        if self.cello_count:
            clef_name = "bass"
            if self.use_percussion_clefs:
                clef_name = "percussion"
            instrument = instruments.Cello()
            instrument_count = self.cello_count
            (
                instrument_staff_group,
                instrument_tag_names,
            ) = self._make_instrument_staff_group(
                clef_name=clef_name,
                count=instrument_count,
                instrument=instrument,
            )
            staff_group.append(instrument_staff_group)
            tag_names.extend(instrument_tag_names)

        ### BASSES ###

        if self.contrabass_count:
            clef_name = "bass_8"
            if self.use_percussion_clefs:
                clef_name = "percussion"
            instrument = instruments.Contrabass()
            instrument_count = self.contrabass_count
            (
                instrument_staff_group,
                instrument_tag_names,
            ) = self._make_instrument_staff_group(
                clef_name=clef_name,
                count=instrument_count,
                instrument=instrument,
            )
            staff_group.append(instrument_staff_group)
            tag_names.extend(instrument_tag_names)

        ### TIME SIGNATURE CONTEXT ###

        global_context = Context(
            lilypond_type="GlobalContext", name="Global_Context", tag=tag
        )
        instrument_tags = " ".join(tag_names)
        tag_string = rf"\tag #'({instrument_tags})"
        literal = LilyPondLiteral(tag_string, "before")
        attach(literal, global_context, tag=tag)
        score.insert(0, global_context)
        return score

    ### PRIVATE METHODS ###

    def _make_instrument_staff_group(self, clef_name=None, count=None, instrument=None):
        site = "abjad.StringOrchestraScoreTemplate._make_instrument_staff_group()"
        tag = Tag(site)
        name = instrument.name.title()
        instrument_staff_group = StaffGroup(
            lilypond_type=f"{name}StaffGroup",
            name=f"{name}_Staff_Group",
            tag=tag,
        )
        tag_names = []
        if count == 1:
            performer_staff_group, tag_name = self._make_performer_staff_group(
                clef_name=clef_name, instrument=instrument, number=None
            )
            instrument_staff_group.append(performer_staff_group)
            tag_names.append(tag_name)
        else:
            for i in range(1, count + 1):
                performer_staff_group, tag_name = self._make_performer_staff_group(
                    clef_name=clef_name, instrument=instrument, number=i
                )
                instrument_staff_group.append(performer_staff_group)
                tag_names.append(tag_name)
        return instrument_staff_group, tag_names

    def _make_performer_staff_group(self, clef_name=None, instrument=None, number=None):
        site = "StringOrchestraScoreTemplate._make_performer_staff_group()"
        tag = Tag(site)
        if number is not None:
            name = f"{instrument.name.title()}_{number}"
        else:
            name = instrument.name.title()
        pitch_range = instrument.pitch_range
        staff_group = StaffGroup(
            lilypond_type="StringPerformerStaffGroup",
            name=f"{name}_Staff_Group",
            tag=tag,
        )
        tag_name = name.replace(" ", "")
        tag_string = rf"\tag #'{tag_name}"
        tag_command = LilyPondLiteral(tag_string, "before")
        attach(tag_command, staff_group, tag=tag)
        if self.split_hands:
            lh_voice = Voice(
                [],
                lilypond_type="FingeringVoice",
                name=f"{name}_Fingering_Voice",
                tag=tag,
            )
            abbreviation = lh_voice.name.lower().replace(" ", "_")
            self.voice_abbreviations[abbreviation] = lh_voice.name
            lh_staff = Staff(
                [lh_voice],
                lilypond_type="FingeringStaff",
                name=f"{name}_Fingering_Staff",
                tag=tag,
            )
            lh_staff.simultaneous = True
            annotate(lh_staff, "pitch_range", pitch_range)
            annotate(lh_staff, "default_clef", Clef(clef_name))
            rh_voice = Voice(
                [],
                lilypond_type="BowingVoice",
                name=f"{name}_Bowing_Voice",
                tag=tag,
            )
            abbreviation = rh_voice.name.lower().replace(" ", "_")
            self.voice_abbreviations[abbreviation] = rh_voice.name
            rh_staff = Staff(
                [rh_voice],
                lilypond_type="BowingStaff",
                name=f"{name}_Bowing_Staff",
                tag=tag,
            )
            rh_staff.simultaneous = True
            staff_group.extend([rh_staff, lh_staff])
        else:
            lh_voice = Voice(
                [],
                lilypond_type="FingeringVoice",
                name=f"{name}_Voice",
                tag=tag,
            )
            lh_staff = Staff(
                [lh_voice],
                lilypond_type="FingeringStaff",
                name=f"{name}_Staff",
                tag=tag,
            )
            lh_staff.simultaneous = True
            annotate(lh_staff, "pitch_range", pitch_range)
            annotate(lh_staff, "default_clef", Clef(clef_name))
            staff_group.append(lh_staff)
        return staff_group, tag_name

    ### PUBLIC PROPERTIES ###

    @property
    def cello_count(self):
        """
        Number of cellos in string orchestra.

        Returns nonnegative integer.
        """
        return self._cello_count

    @property
    def contrabass_count(self):
        """
        Number of contrabasses in string orchestra.

        Returns nonnegative integer.
        """
        return self._contrabass_count

    @property
    def split_hands(self):
        """
        Is true if each performer's hand receives a separate staff.
        """
        return self._split_hands

    @property
    def use_percussion_clefs(self):
        """
        Is true if each staff should use a percussion clef rather than the
        normal clef for that instrument.
        """
        return self._use_percussion_clefs

    @property
    def viola_count(self):
        """
        Number of violas in string orcestra.

        Returns nonnegative integer.
        """
        return self._viola_count

    @property
    def violin_count(self):
        """
        Number of violins in string orchestra.

        Returns nonnegative integer.
        """
        return self._violin_count


class StringQuartetScoreTemplate(ScoreTemplate):
    r"""
    String quartet score template.

    ..  container:: example

        >>> template = abjad.StringQuartetScoreTemplate()
        >>> abjad.show(template) # doctest: +SKIP

        >>> string = abjad.lilypond(template.__illustrate__()[abjad.Score])
        >>> print(string)
        \context Score = "String_Quartet_Score"
        <<
            \context StaffGroup = "String_Quartet_Staff_Group"
            <<
                \tag #'first-violin
                \context Staff = "First_Violin_Staff"
                {
                    \context Voice = "First_Violin_Voice"
                    {
                        \clef "treble"
                        s1
                    }
                }
                \tag #'second-violin
                \context Staff = "Second_Violin_Staff"
                {
                    \context Voice = "Second_Violin_Voice"
                    {
                        \clef "treble"
                        s1
                    }
                }
                \tag #'viola
                \context Staff = "Viola_Staff"
                {
                    \context Voice = "Viola_Voice"
                    {
                        \clef "alto"
                        s1
                    }
                }
                \tag #'cello
                \context Staff = "Cello_Staff"
                {
                    \context Voice = "Cello_Voice"
                    {
                        \clef "bass"
                        s1
                    }
                }
            >>
        >>

    Returns score template.
    """

    ### CLASS VARIABLES ###

    __slots__ = ()

    ### INITIALIZER ###

    def __init__(self):
        super().__init__()
        self.voice_abbreviations.update(
            {
                "vn1": "First Violin Voice",
                "vn2": "Second Violin Voice",
                "va": "Viola Voice",
                "vc": "Cello Voice",
            }
        )

    ### SPECIAL METHODS ###

    def __call__(self):
        """
        Calls string quartet score template.

        Returns score.
        """
        site = "abjad.StringQuartetScoreTemplate.__call__()"
        tag = Tag(site)

        # make first violin voice and staff
        first_violin_voice = Voice([], name="First_Violin_Voice", tag=tag)
        first_violin_staff = Staff(
            [first_violin_voice], name="First_Violin_Staff", tag=tag
        )
        clef = Clef("treble")
        annotate(first_violin_staff, "default_clef", clef)
        violin = instruments.Violin()
        annotate(first_violin_staff, "default_instrument", violin)
        literal = LilyPondLiteral(r"\tag #'first-violin", "before")
        attach(literal, first_violin_staff)

        # make second violin voice and staff
        second_violin_voice = Voice([], name="Second_Violin_Voice", tag=tag)
        second_violin_staff = Staff(
            [second_violin_voice], name="Second_Violin_Staff", tag=tag
        )
        clef = Clef("treble")
        annotate(second_violin_staff, "default_clef", clef)
        violin = instruments.Violin()
        annotate(second_violin_staff, "default_instrument", violin)
        literal = LilyPondLiteral(r"\tag #'second-violin", "before")
        attach(literal, second_violin_staff)

        # make viola voice and staff
        viola_voice = Voice([], name="Viola_Voice", tag=tag)
        viola_staff = Staff([viola_voice], name="Viola_Staff", tag=tag)
        clef = Clef("alto")
        annotate(viola_staff, "default_clef", clef)
        viola = instruments.Viola()
        annotate(viola_staff, "default_instrument", viola)
        literal = LilyPondLiteral(r"\tag #'viola", "before")
        attach(literal, viola_staff)

        # make cello voice and staff
        cello_voice = Voice([], name="Cello_Voice", tag=tag)
        cello_staff = Staff([cello_voice], name="Cello_Staff", tag=tag)
        clef = Clef("bass")
        annotate(cello_staff, "default_clef", clef)
        cello = instruments.Cello()
        annotate(cello_staff, "default_instrument", cello)
        literal = LilyPondLiteral(r"\tag #'cello", "before")
        attach(literal, cello_staff)

        # make string quartet staff group
        string_quartet_staff_group = StaffGroup(
            [first_violin_staff, second_violin_staff, viola_staff, cello_staff],
            name="String_Quartet_Staff_Group",
            tag=tag,
        )

        # make string quartet score
        string_quartet_score = Score(
            [string_quartet_staff_group],
            name="String_Quartet_Score",
            tag=tag,
        )

        # return string quartet score
        return string_quartet_score


class TwoStaffPianoScoreTemplate(ScoreTemplate):
    r"""
    Two-staff piano score template.

    ..  container:: example

        >>> template = abjad.TwoStaffPianoScoreTemplate()
        >>> abjad.show(template) # doctest: +SKIP

        >>> string = abjad.lilypond(template.__illustrate__()[abjad.Score])
        >>> print(string)
        \context Score = "Two_Staff_Piano_Score"
        <<
            \context GlobalContext = "Global_Context"
            <<
                \context GlobalRests = "Global_Rests"
                {
                }
                \context GlobalSkips = "Global_Skips"
                {
                }
            >>
            \context PianoStaff = "Piano_Staff"
            <<
                \context Staff = "RH_Staff"
                {
                    \context Voice = "RH_Voice"
                    {
                        s1
                    }
                }
                \context Staff = "LH_Staff"
                {
                    \context Voice = "LH_Voice"
                    {
                        \clef "bass"
                        s1
                    }
                }
            >>
        >>

    Returns score template.
    """

    ### CLASS VARIABLES ###

    __slots__ = ()

    ### INITIALIZER ###

    def __init__(self):
        super().__init__()
        self.voice_abbreviations.update({"rh": "RHVoice", "lh": "LHVoice"})

    ### SPECIAL METHODS ###

    def __call__(self) -> Score:
        r"""
        Calls two-staff piano score template.

        ..  container:: example

            REGRESSION. Attaches piano to piano group (rather than just staff):

            >>> template = abjad.TwoStaffPianoScoreTemplate()
            >>> score = template()
            >>> piano_staff = score['Piano_Staff']
            >>> rh_voice = score['RH_Voice']
            >>> lh_voice = score['LH_Voice']

            >>> rh_voice.append("g'4")
            >>> lh_voice.append("c4")

            >>> abjad.show(score) # doctest: +SKIP

            ..  docs::

                >>> string = abjad.lilypond(score)
                >>> print(string)
                \context Score = "Two_Staff_Piano_Score"
                <<
                    \context GlobalContext = "Global_Context"
                    <<
                        \context GlobalRests = "Global_Rests"
                        {
                        }
                        \context GlobalSkips = "Global_Skips"
                        {
                        }
                    >>
                    \context PianoStaff = "Piano_Staff"
                    <<
                        \context Staff = "RH_Staff"
                        {
                            \context Voice = "RH_Voice"
                            {
                                g'4
                            }
                        }
                        \context Staff = "LH_Staff"
                        {
                            \context Voice = "LH_Voice"
                            {
                                c4
                            }
                        }
                    >>
                >>

        """
        site = "abjad.TwoStaffPianoScoreTemplate.__call__()"
        tag = Tag(site)
        # GLOBAL CONTEXT
        global_context = self._make_global_context()

        # RH STAFF
        rh_voice = Voice(name="RH_Voice", tag=tag)
        rh_staff = Staff([rh_voice], name="RH_Staff", tag=tag)

        # LH STAFF
        lh_voice = Voice(name="LH_Voice", tag=tag)
        lh_staff = Staff([lh_voice], name="LH_Staff", tag=tag)
        annotate(lh_staff, "default_clef", Clef("bass"))

        # PIANO STAFF
        staff_group = StaffGroup(
            [rh_staff, lh_staff],
            lilypond_type="PianoStaff",
            name="Piano_Staff",
            tag=tag,
        )
        annotate(staff_group, "default_instrument", Piano())

        # SCORE
        score = Score(
            [global_context, staff_group],
            name="Two_Staff_Piano_Score",
            tag=tag,
        )
        return score



import collections
import typing

from . import _inspect, score
from .indicators.RepeatTie import RepeatTie
from .indicators.Tie import Tie


def _coerce_exclude(exclude):
    if exclude is None:
        exclude = ()
    elif isinstance(exclude, str):
        exclude = (exclude,)
    else:
        exclude = tuple(exclude)
    assert isinstance(exclude, tuple), repr(exclude)
    return exclude


def _is_unpitched(LEAF):
    if hasattr(LEAF, "written_pitch"):
        return False
    if hasattr(LEAF, "written_pitches"):
        return False
    return True


def _get_logical_tie_leaves(LEAF):
    if _is_unpitched(LEAF):
        return [LEAF]
    leaves_before, leaves_after = [], []
    current_leaf = LEAF
    while True:
        previous_leaf = _get_leaf(current_leaf, -1)
        if previous_leaf is None:
            break
        if _is_unpitched(previous_leaf):
            break
        if current_leaf._has_indicator(RepeatTie) or previous_leaf._has_indicator(Tie):
            leaves_before.insert(0, previous_leaf)
        else:
            break
        current_leaf = previous_leaf
    current_leaf = LEAF
    while True:
        next_leaf = _get_leaf(current_leaf, 1)
        if next_leaf is None:
            break
        if _is_unpitched(next_leaf):
            break
        if current_leaf._has_indicator(Tie) or next_leaf._has_indicator(RepeatTie):
            leaves_after.append(next_leaf)
        else:
            break
        current_leaf = next_leaf
    leaves = leaves_before + [LEAF] + leaves_after
    return leaves


def _iterate_components(
    client,
    prototype=None,
    *,
    exclude=None,
    do_not_iterate_grace_containers=None,
    grace=None,
    reverse=None,
):
    argument = client
    prototype = prototype or score.Component
    before_grace_container = None
    after_grace_container = None
    exclude = _coerce_exclude(exclude)
    assert isinstance(exclude, tuple), repr(exclude)
    if grace is not False and isinstance(argument, score.Leaf):
        before_grace_container = argument._before_grace_container
        after_grace_container = argument._after_grace_container
    if not reverse:
        if (
            not do_not_iterate_grace_containers
            and grace is not False
            and before_grace_container
        ):
            yield from _iterate_components(
                before_grace_container,
                prototype,
                do_not_iterate_grace_containers=do_not_iterate_grace_containers,
                grace=grace,
                reverse=reverse,
            )
        if isinstance(argument, prototype):
            if (
                grace is None
                or (grace is True and _inspect._get_grace_container(argument))
                or (grace is False and not _inspect._get_grace_container(argument))
            ):
                if not _should_exclude(argument, exclude):
                    yield argument
        if (
            not do_not_iterate_grace_containers
            and grace is not False
            and after_grace_container
        ):
            yield from _iterate_components(
                after_grace_container,
                prototype,
                exclude=exclude,
                do_not_iterate_grace_containers=do_not_iterate_grace_containers,
                grace=grace,
                reverse=reverse,
            )
        if isinstance(argument, collections.abc.Iterable):
            for item in argument:
                yield from _iterate_components(
                    item,
                    prototype,
                    exclude=exclude,
                    do_not_iterate_grace_containers=do_not_iterate_grace_containers,
                    grace=grace,
                    reverse=reverse,
                )
    else:
        if (
            not do_not_iterate_grace_containers
            and grace is not False
            and after_grace_container
        ):
            yield from _iterate_components(
                after_grace_container,
                prototype,
                exclude=exclude,
                do_not_iterate_grace_containers=do_not_iterate_grace_containers,
                grace=grace,
                reverse=reverse,
            )
        if isinstance(argument, prototype):
            if (
                grace is None
                or (grace is True and _inspect._get_grace_container(argument))
                or (grace is False and not _inspect._get_grace_container(argument))
            ):
                if not _should_exclude(argument, exclude):
                    yield argument
        if (
            not do_not_iterate_grace_containers
            and grace is not False
            and before_grace_container
        ):
            yield from _iterate_components(
                before_grace_container,
                prototype,
                exclude=exclude,
                do_not_iterate_grace_containers=do_not_iterate_grace_containers,
                grace=grace,
                reverse=reverse,
            )
        if isinstance(argument, collections.abc.Iterable):
            for item in reversed(argument):
                yield from _iterate_components(
                    item,
                    prototype,
                    exclude=exclude,
                    do_not_iterate_grace_containers=do_not_iterate_grace_containers,
                    grace=grace,
                    reverse=reverse,
                )


def _iterate_descendants(component, cross_offset=None):
    assert isinstance(component, score.Component), repr(component)
    if component is None:
        components = ()
    else:
        generator = _public_iterate_components(component)
        components = list(generator)
    result = []
    if cross_offset is None:
        result = components
    else:
        for component in components:
            append_x = True
            if not (
                component._get_timespan().start_offset < cross_offset
                and cross_offset < component._get_timespan().stop_offset
            ):
                append_x = False
            if append_x:
                result.append(component)
    return tuple(result)


def _iterate_logical_ties(
    ARGUMENT,
    *,
    exclude=None,
    grace=None,
    nontrivial=None,
    pitched=None,
    reverse=None,
    wrapper_class=None,
) -> typing.Generator:
    yielded_logical_ties = set()
    for leaf in _public_iterate_leaves(
        ARGUMENT, exclude=exclude, grace=grace, pitched=pitched, reverse=reverse
    ):
        leaves = _get_logical_tie_leaves(leaf)
        # logical_tie = LogicalTie(leaves)
        # if leaf is not logical_tie.head:
        if leaf is not leaves[0]:
            continue
        if (
            nontrivial is None
            or (nontrivial is True and not len(leaves) == 1)
            or (nontrivial is False and len(leaves) == 1)
        ):
            # if logical_tie not in yielded_logical_ties:
            if wrapper_class is not None:
                leaves = wrapper_class(leaves)
            if leaves not in yielded_logical_ties:
                # yielded_logical_ties.add(logical_tie)
                yielded_logical_ties.add(leaves)
                # yield logical_tie
                yield leaves


def _get_leaf(ARGUMENT, n: int = 0):
    if n not in (-1, 0, 1):
        message = "n must be -1, 0 or 1:\n"
        message += f"   {repr(n)}"
        raise Exception(message)
    if isinstance(ARGUMENT, score.Leaf):
        candidate = _inspect._get_sibling_with_graces(ARGUMENT, n)
        if isinstance(candidate, score.Leaf):
            return candidate
        return _inspect._get_leaf_from_leaf(ARGUMENT, n)
    if 0 <= n:
        reverse = False
    else:
        reverse = True
        n = abs(n) - 1
    leaves = _public_iterate_leaves(ARGUMENT, reverse=reverse)
    for i, leaf in enumerate(leaves):
        if i == n:
            return leaf
    return None


def _public_iterate_components(
    ARGUMENT, prototype=None, *, exclude=None, grace=None, reverse=None
):
    if isinstance(ARGUMENT, score.Container):
        for component in _iterate_components(
            ARGUMENT,
            prototype,
            exclude=exclude,
            do_not_iterate_grace_containers=False,
            grace=grace,
            reverse=reverse,
        ):
            yield component
    elif isinstance(ARGUMENT, collections.abc.Iterable):
        if not reverse:
            for item in ARGUMENT:
                generator = _public_iterate_components(
                    item,
                    prototype,
                    exclude=exclude,
                    grace=grace,
                    reverse=reverse,
                )
                yield from generator
        else:
            for item in reversed(ARGUMENT):
                generator = _public_iterate_components(
                    item,
                    prototype,
                    exclude=exclude,
                    grace=grace,
                    reverse=reverse,
                )
                yield from generator
    else:
        for component in _iterate_components(
            ARGUMENT,
            prototype,
            exclude=exclude,
            do_not_iterate_grace_containers=True,
            grace=grace,
            reverse=reverse,
        ):
            yield component


def _public_iterate_leaves(
    ARGUMENT,
    prototype=None,
    *,
    exclude=None,
    grace=None,
    pitched=None,
    reverse=None,
):
    prototype = prototype or score.Leaf
    if pitched is True:
        prototype = (score.Chord, score.Note)
    elif pitched is False:
        prototype = (score.MultimeasureRest, score.Rest, score.Skip)
    return _public_iterate_components(
        ARGUMENT, prototype=prototype, exclude=exclude, grace=grace, reverse=reverse
    )


def _should_exclude(argument, exclude):
    assert isinstance(exclude, tuple)
    for string in exclude:
        if argument._has_indicator(string):
            return True
    return False
